import os
import torch

module_type = type(os)

def _g():
    '''used only for getting generator_type'''
    yield from range(1)
generator_type =  type(_g())

def argparse_list_type(element_type=int):
    '''used to put into argparse add_argument(type=?),
    in order to parse lists like:
    --gpu-ids=1,2,3,4

    usage:
        floatlist = argparse_list_type(float)
        parser.add_argument('--cuda-devs',type=flaotlist,default=[1,2])

    args:
        element_type: type of each element in the list,
                eache element will be cat to this type.

    Notice: Instances must be used, instead of this class itself.
    '''
    def list_type(option_string):
        a = option_string
        a = a.strip().strip('[]()\{\}')
        sep = ' '
        if ',' in a:
            sep = ','
        elems = [i.strip() for i in a.split(sep)]

        return [element_type(i) for i in elems]
    return list_type

list_of_int = argparse_list_type()

list_of_float = argparse_list_type(float)

def argparse_bool_type(option_string):
    '''used to put into argparse add_argument(type=torchfun.bool),
    in order to parse bool switch values like:
    false False true True 0 1'''
    a = option_string.strip().lower()
    if a in ('true','1'):
        return True
    elif a in ('false','0'):
        return False
    else:
        raise Exception('un-parsable argument value %s for bool-type'%option_string)

bool=argparse_bool_type
Bool=argparse_bool_type

class TorchEasyList(list):
    def cat(self,dim=0):
        return torch.cat(self,dim)
    def push_head(self,element):
        self.insert(0,element)
    def push_back(self,element):
        return self.append(element)
    def __add__(self,element):
        if isinstance(element,list):
            return TorchEasyList(super().__add__(element))
        else:
            newlist = TorchEasyList(self)
            newlist.push_head(element)
            return newlist



__doc__='''tools for user-interface interactions. 
(mostly for console-debugging)'''

def info(*args):
    header = '[info]\t'
    print(header,*args)
    
def error(*args):
    header = '[error]\t'
    print(header,*args)
    exit(-1)

def warning(*args):
    header = '[warn]\t'
    print(header,*args)
warn = warning

def wait(*args):
    header = '[wait]\t'
    print(header,*args)
    print('\t\tpress Enter to continue')
    input()

def input_or(prompt,default='y'):
    v = input(prompt+'(default:%s):'%(str(default)))
    v = v or default
    dtype = type(default)
    return dtype(v)



from __future__ import division,print_function
import torch
from .torchfun import sort_args
import torch.nn.functional as F
import itertools

__doc__ = 'mathematical functions'

def conv2d_dfs(x,weight,bias=None,stride=1,padding=0,dilation=1):
    '''depth fully shared conv2d.
    Argument:
        x: input image with size: N x C x H x W
        weight: shape shoule be : 1 x 1 x kernel-height x k-width
        bias: contains only 1 number or None
    '''
    b,c,h,w = x.shape
    x = x.contiguous().view(b*c,1,h,w)
    x = F.conv2d(x,weight,bias,stride,padding,dilation,groups=1)
    _,_,h,w = x.shape
    return x.contiguous().view(b,c,h,w)

def flatten(x):
    '''Flatten function
    Usage:
        out = flatten(x)
    '''
    #shapes = x.shape
    #dims = len(shapes)
    #flatten_length = torch.prod(shapes[1:])
    return x.contiguous().view(x.size(0),-1)

def subpixel(x,out_channels=1):
    '''Unfold channel/depth dimensions to enlarge the feature map
    Notice: Output size is deducted. 
    The size of the unfold square is automatically determined
    e.g. :
        images: 100x9x16x16.  9=3x3 square
        subpixel-out: 100x1x48x48
    Arguement:
        x: NCHW image, channel first.
        out_channels, channel number of output feature map'''
    b,c,h,w = shapes = list(x.shape)
    out_c = int(out_channels)
    if c%out_c != 0:
        print('input has',c,'channels, cannot be split into',out_c,'parts')
        raise Exception('subpixel inappropriate size')
    unfold_dim = int(c//out_c)
    l = int(unfold_dim**0.5)
    if l**2 != unfold_dim:
        print('remaining',unfold_dim,'digits for each channel, unable to sqrt.')
        raise Exception('subpixel inappropriate size %f**2 != %d' %(l,int(unfold_dim)))
    ###### start ######
    x = x.transpose(1,3) # nhwc
    y = x.reshape(b,h,w,unfold_dim,out_c)
    y = y.transpose(2,3)
    y = y.reshape(b,h,l,l,w,out_c)
    y = y.reshape(b,l*h,l,w,out_c) # b,l*h,l,w,outc
    y = y.transpose(2,3) # b,l*h,w,l,outc
    y = y.reshape(b,l*h,l*w,out_c)
    y = y.transpose(1,3) # nchw
    return y

def clip(in_tensor,max_or_min,min_or_max):
    '''limit the values in in_tensor to be within [min,max].
    values larger than max will be cut to max, respectively for mins.
    the order of max/min doesn't matter.
    the operation is not in-place, that saves you alot troubles.
    Notice: this clip is not inplace, and neither can pass backward derivatives
            when the values are clipped to min/max.
    Warning: When applied during loss-calculating in training, this clip will cause
            gradient disapperance.
    '''
    minv,maxv = torch.tensor(sorted([max_or_min,min_or_max]),dtype=in_tensor.dtype).to(in_tensor.device)
    x = torch.max(in_tensor,minv)
    x = torch.min(x,maxv)
    return x

def clip_(in_tensor,max_or_min,min_or_max):
    '''limit the values in in_tensor to be within [min,max].
    values larger than max will be cut to max, respectively for mins.
    the order of max/min doesn't matter.
    the operation is not in-place, that saves you alot troubles.
    Notice: this clip is not inplace.
            But, for clipped values, the gradients will always be passed backwards.
            This is useful when 
    '''
    minv,maxv = sorted((max_or_min,min_or_max))

    greater_pos = (in_tensor.detach()>maxv).type_as(in_tensor)
    diff = (maxv-in_tensor.detach())*greater_pos
    in_tensor = in_tensor+diff

    less_pos = (in_tensor.detach()<minv).type_as(in_tensor)
    diff = (minv-in_tensor.detach())*less_pos
    in_tensor = in_tensor+diff
    return in_tensor



def add_noise(in_tensor,noise_type='normal',noise_param=(0,1),range_limit=(-1,1)):
    '''
    Add noise to input tensor.
    Noise type can be either `normal` or `uniform`
        * for normal, (mean,std) is required as noise_param
        * for uniform (min,max) is required as noise_param
    The range of the output tensor can be limited,
      by giving `range_limit`:(min,max)
    '''
    if noise_type=='uniform':
        nmin,nmax = noise_param
        noise = torch.zeros_like(in_tensor).uniform_(nmin,nmax)
    elif noise_type=='normal':
        mean,std = noise_param
        noise = torch.zeros_like(in_tensor).normal_(mean,std)
    if range_limit is not None:
        return clip(in_tensor+noise,*range_limit)
    else:
        return in_tensor+noise


def instance_mean_std(x,num_features):
    '''NCHW'''
    b,c,h,w = x.shape
    nc_groups = c//num_features
    x = x.contiguous().view(b,nc_groups,num_features,-1)
    x = x.transpose(1,2) # b, numfeatu, ncgroup, h*w
    x = x.contiguous()
    x = x.contiguous().view(b*num_features,-1)

    mean = x.mean(1)
    std = x.std(1)

    mean = mean.contiguous().view(b,num_features) # n,feature
    std = std.contiguous().view(b,num_features) # n,feature
    return mean,std

def instance_renorm(x,mean,std,eps=1e-5):
    '''Make x have given mean and std.
    Argument:
    x: NCHW
    mean: tensor with size: N-by-num-features
    std: tensor with size: N-by-num-features
    eps: default 1e-5
    '''
    b,c,h,w = x.shape
    b,num_features = mean.shape
    nc_groups = c//num_features

    x_m,x_std = instance_mean_std(x,num_features)
    x_m,x_std = x_m.detach(),x_std.detach()
    x = x.contiguous().view(b,nc_groups,num_features,-1)
    x = x.transpose(2,3).transpose(0,2) # -1,ncgroup,b,numfeature
    x = ((x-x_m)/(x_std+eps))*(std+eps) + mean
    x = x.transpose(0,2).transpose(2,3) # b,nc_groups,num_features,-1
    x = x.contiguous().view(b,c,h,w)
    return x.contiguous()

def max_min_norm(x,to_max,to_min,eps=1e-5):
    '''scale the input x into range [to_min,to_max].
    Arguments:
        to_max: target expect max value of the output
        to_min: target expect min value of the output
    '''
    b,c,h,w = x.shape
    x = flatten(x)
    xmax,_ = x.max(1)
    xmin,_ = x.min(1)
    x = x.transpose(0,1)
    x = (x-xmin)/(xmax-xmin+eps)
    x = x * (to_max - to_min) + to_min
    x = x.transpose(0,1)
    return x.view(b,c,h,w)

### training related functional tools
def combine_parameters(*models):
    '''Combine the parameters of serveral trainable module object,
    into one unified parameter generator.
    Arguements:
            *models: any number of models.

    This utility is useful when you want several individual parts to be
    handled by one Optimizer. Parameters shall be gathered into one iterator
    first, because torch.optimizers require only one parameter-iterator as input'''
    return itertools.chain(*[m.parameters() for m in models])

def cosine_similarity(x1,x2):
    '''compute cosine similarity between two batch of data.
    x1 x2 are flattened first, into N x V shape.
    then each paired sample between x1 and x2 are used to compute a 
    multi-dimensional vectorial cosine similarity.

    output:  similarity output with size N, which is a vector.'''
    x1=flatten(x1)
    x2=flatten(x2)
    return torch.nn.functional.cosine_similarity(x1,x2,dim=1)


import torch
import torch.utils.data as data
from PIL import Image
import os
import numpy as np
from torchvision.datasets.folder import find_classes,make_dataset
from torchvision.datasets.folder import DatasetFolder,ImageFolder
from torchvision.datasets.folder import default_loader,IMG_EXTENSIONS

__doc__ = 'provide dataset related classes'


class ApposeDataset(torch.utils.data.Dataset):
    '''make several dataset abreast, 
    returning tuples of their corresponding elements.
    Usage:
            dataset = ApposeDataset(dataset1,dataset2,...)'''
    def __init__(self,*datasets):
        super().__init__()
        self.datasets=datasets
        self.check_consistency()
    def check_consistency(self):
        lens=[len(d) for d in self.datasets]
        min_length=min(lens)
        max_length=max(lens)
        if min_length!=max_length:
            print('''torchfun:ApposeDataset:warning:input datasets are of different size, 
                the minimal length will be used to access each dataset.''')
        self.length = min_length
        self.lens = lens
    def __len__(self):
        return self.length
    def __getitem__(self,index):
        if index>=self.length:
            raise IndexError()
        else:
            return (d[index] for d in self.datasets)


class ImageAugmentationDataset(ImageFolder):
    '''
        Pre_transform is applied to the source image firstly.
        Then, degrade_transform is applied to get degraded image (such as blurred image).
        Finally, both the degraded imgs and the pre-transformed images are uniformly post-transformed.
        Usually, the post-transforms are ToTensor() and Normalize()

        Degrade transform should have two arguments as input:
            1: input image to be degraded
            2: Boolean input indicates if the degrading parameters should be returned.
    '''
    def __init__(self, root, 
                pre_transform=None, 
                degrade_transform=None,
                post_transform=None,
                loader=default_loader):
        super(self.__class__, self).__init__(root, 
                            transform=None,
                            target_transform=None,
                            loader=loader)
        self.imgs = self.samples
        self.pre_transform = pre_transform
        self.degrade_transform = degrade_transform
        self.post_transform = post_transform        


    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (sample, target) where target is class_index of the target class.
        """
        path, target = self.samples[index]
        sample = self.loader(path)
        if self.pre_transform:
            sample = self.pre_transform(sample)
        degraded_sample = sample.copy()
        if self.degrade_transform:
            degraded_sample = self.degrade_transform(degraded_sample)

        if self.post_transform:
            sample = self.post_transform(sample)
            degraded_sample = self.post_transform(degraded_sample)

        return degraded_sample,sample



class BatchQueue(object):
    ## TODO: modify for torch api
    def __init__(self, max_length=50, y_max=1.5, y_min=0.8, 
        img_size=128,
        channels=3):
        '''
            if max length is 50, then there are at most 25 real samples and 25 fake
            sample in the queue.
        '''
        self.real=[]
        self.fake=[]
        self.max_length=max_length//2
        self.real_generator=None
        self.y_max = y_max
        self.y_min = y_min
        self.width = img_size
        self.height = img_size
        self.channels = channels
        self.input_shape = (self.height, self.width, self.channels)
        self.real_generator=None
        self.base_generator=None
        self.faking_model=None

    def init_with_generator(self,real_generator, base_generator, faking_model):
        '''
        real will be marked as 1, 
        fake samples generated by base_generator will be marked as 0
        '''
        self.real_generator=real_generator
        self.base_generator=base_generator
        self.faking_model=faking_model
        for i in range(self.max_length):
            real = get_image_batch(real_generator)
            fake = faking_model.predict(get_image_batch(base_generator))
            self.push_real(real)
            self.push_fake(fake)
        return self

    def init_with_data(self,real,fake):
        self.push_real(real)
        self.push_fake(fake)

    def reinit(self):
        '''
        init the queue again after parameters are loaded by restore()
        '''
        return self.init(self.real_generator,self.base_generator,self.faking_model)

    def push(self,real,fake):
        self.push_real(real)
        self.push_fake(fake)

    def push_real(self,realimg):
        return self._push(self.real,realimg)

    def push_fake(self,fakeimg):
        return self._push(self.fake,fakeimg)    

    def _push(self, queue, imgs):
        shape = imgs.shape
        # batch of images is pushing in
        batch_size = shape[0]
        if shape[1:] != self.input_shape:
            print('BatchQueue[Warning]:input',shape,'inconsistent with',self.input_shape)
            #return False
        for i in range(batch_size):
            self.__push(queue,imgs[i:i+1,:,:,:])
        return True

    def __push(self, queue, img):
        if len(queue)< self.max_length:
            queue.insert(0,img)
        else:
            queue.pop()
            self._push(queue,img)
 
    def get(self,queue):
        '''
        input: List[np.vector]
        make an np.ndarray with size of: all x columns
        '''
        l=len(queue)
        dat=np.concatenate(queue,0)
        return dat,l

    def get_real(self):
        r,_ = self.get(self.real)
        return r
    def get_fake(self):
        r,_ = self.get(self.fake)
        return r

    def get_batch(self):
        '''
        get a batch with batch size of self.max_length*2
        '''
        realbatch, realbatchsize = self.get(self.real)
        fakebatch, fakebatchsize = self.get(self.fake)
        x=np.concatenate([realbatch,fakebatch],0)
        ones=np.random.uniform(low=self.y_min,high=self.y_max,size=realbatchsize)
        zeros=np.random.uniform(low=-self.y_max,high=-self.y_min,size=fakebatchsize)
        y=np.concatenate([ones,zeros],0)
        return x,y

    def get_real_fake(self):
        r,_ = self.get(self.real)
        f,_ = self.get(self.fake)
        return r,f

    def new_sample(self):
        '''
        returns one img or imgs according to generator's batch size configuration
        '''
        imgs = get_image_batch(self.real_generator) # ranges -1,1
        return imgs, np.random.uniform(low=self.y_min,high=self.y_max,size=imgs.shape[0])



# LICENSE: MIT
# Author: CHEN Si Yu.
# Date: 2018
# Appended constrains: If the content in this project is used in your source-codes, this author's name must be cited at the begining of your source-code. 

from __future__ import division
from PIL import ImageFilter
import torch
import math
import random
from PIL import Image, ImageOps, ImageEnhance
try:
    import accimage
except ImportError:
    accimage = None
import numpy as np
import numbers
import types
import collections
import warnings

from torchvision.transforms import functional as F

__all__ = [
"Transform",
"RandomGaussianBlur",
"NoTransform",
"Normalize",
"Resize",
"CenterCrop",
"RandomCrop",
"RandomHorizontalFlip",
"RandomVerticalFlip",
"RandomResizedCrop",
"FiveCrop",
"TenCrop",
"LinearTransformation",
"ColorJitter",
"RandomRotation",
"RandomAffine",
"Grayscale",
"RandomGrayscale"
]

_pil_interpolation_to_str = {
    Image.NEAREST: 'PIL.Image.NEAREST',
    Image.BILINEAR: 'PIL.Image.BILINEAR',
    Image.BICUBIC: 'PIL.Image.BICUBIC',
    Image.LANCZOS: 'PIL.Image.LANCZOS',
}



__doc__ = '''Transforms used for data augmentation in torchvision.datasets.*
or in torch.utils.data.Dataset

New function: most of the transforms support x-y paired transformation.
New function: a place-holder base class Transform is created and used as the 
    base class of all transforms. 
    This change is made so that more complicated metatype-programming 
    can be possible.

Some of the code is borrowed from an old project but the original author is not found.

Please let me know if any part of this code is from your contribution.'''

class Transform:
    pass

class RandomGaussianBlur(Transform):
    '''PIL image'''
    def __init__(self,kernel_ratio=0.01,random_ratio=0.005,pixel_range=None):
        '''
        Arguemet:
            kernel_ratio: radius ratio with respect to image width
            random_ratio: random_noise applied to kernel_ratio
        '''
        super(self.__class__,self).__init__()
        self.__dict__.update(locals())
        assert(random_ratio<kernel_ratio)

    def __call__(self,img, target=None):
        l,w = img.size
        avg_size = (l+w)//2

        kernel_size_ratio = (random.random()-0.5)*2*self.random_ratio \
                    + self.kernel_ratio
        kernel_size = max(int(kernel_size_ratio * avg_size),1)
        fltr = ImageFilter.GaussianBlur(kernel_size)
        if target is None:
            return img.filter(fltr)
        else:
            return img.filter(fltr),target

    def __repr__(self):
        return 'GaussianBlur with kernel_ratio:%f random_ratio:%f' %(self.kernel_ratio,self.random_ratio)
        
class NoTransform(Transform):
    def __init__(self,*rubbish_args,**rubbish_kws):
        super().__init__()
    def __call__(self,img,target=None):
        if target is not None:
            return img,target
        else:
            return img
    def __repr__(self):
        return 'No Transform:'+super().__repr__()


class Normalize(Transform):
    """Normalize a tensor image with mean and standard deviation.
    Given mean: ``(M1,...,Mn)`` and std: ``(S1,..,Sn)`` for ``n`` channels, this transform
    will normalize each channel of the input ``torch.*Tensor`` i.e.
    ``input[channel] = (input[channel] - mean[channel]) / std[channel]``

    Args:
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channel.
    """

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor1, tensor2=None):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.

        Returns:
            Tensor: Normalized Tensor image.
        """
        if tensor2 is not None:
            return F.normalize(tensor1, self.mean, self.std), F.normalize(tensor2, self.mean, self.std)
        else:
            return F.normalize(tensor1, self.mean, self.std)

    def __repr__(self):
        return self.__class__.__name__ + '(mean={0}, std={1})'.format(self.mean, self.std)


class Resize(Transform):
    """Resize the input PIL Image to the given size.

    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (h, w), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``
        interpolation_tg (int, optional): Desired interpolation for target. Default is
            ``PIL.Image.NEAREST``
    """

    def __init__(self, size, interpolation=Image.BILINEAR, interpolation_tg = Image.NEAREST):
        assert isinstance(size, int) or (isinstance(size, collections.Iterable) and len(size) == 2)
        self.size = size
        self.interpolation = interpolation
        self.interpolation_tg = interpolation_tg

    def __call__(self, img, target = None):
        """
        Args:
            img (PIL Image): Image to be scaled.
            target (PIL Image): (optional) Target to be scaled

        Returns:
            PIL Image: Rescaled image(s).
        """
        if target is not None:
            return F.resize(img, self.size, self.interpolation), F.resize(target, self.size, self.interpolation_tg)
        return F.resize(img, self.size, self.interpolation)

    def __repr__(self):
        interpolate_str = _pil_interpolation_to_str[self.interpolation]
        return self.__class__.__name__ + '(size={0}, interpolation={1})'.format(self.size, interpolate_str)

class CenterCrop(Transform):
    """Crops the given PIL Image at the center.

    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, img, target=None):
        """
        Args:
            img (PIL Image): Image to be cropped.
            target (PIL Image): (optional) Target to be cropped

        Returns:
            PIL Image: Cropped image(s).
        """
        if target is not None:
            return F.center_crop(img, self.size), F.center_crop(target, self.size)
        return F.center_crop(img, self.size)

    def __repr__(self):
        return self.__class__.__name__ + '(size={0})'.format(self.size)


class RandomCrop(Transform):
    """Crop the given PIL Image at a random location.

    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
        padding (int or sequence, optional): Optional padding on each border
            of the image. Default is 0, i.e no padding. If a sequence of length
            4 is provided, it is used to pad left, top, right, bottom borders
            respectively.
        pad_if_needed (boolean): It will pad the image if smaller than the
            desired size to avoid raising an exception.
    """

    def __init__(self, size, padding=0, pad_if_needed=False):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.padding = padding
        self.pad_if_needed = pad_if_needed

    @staticmethod
    def get_params(img, output_size):
        """Get parameters for ``crop`` for a random crop.

        Args:
            img (PIL Image): Image to be cropped.
            output_size (tuple): Expected output size of the crop.

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for random crop.
        """
        w, h = img.size
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def __call__(self, img, target = None):
        """
        Args:
            img (PIL Image): Image to be cropped.
            target (PIL Image): (optional) Target to be cropped

        Returns:
            PIL Images: Cropped image(s).
        """
        if self.padding > 0:
            img = F.pad(img, self.padding)
            if target is not None:
                target = F.pad(target, self.padding)

        # pad the width if needed
        if self.pad_if_needed and img.size[0] < self.size[1]:
            img = F.pad(img, (int((1 + self.size[1] - img.size[0]) / 2), 0))
            if target is not None:
                target = F.pad(target, (int((1 + self.size[1] - target.size[0]) / 2), 0))
        # pad the height if needed
        if self.pad_if_needed and img.size[1] < self.size[0]:
            img = F.pad(img, (0, int((1 + self.size[0] - img.size[1]) / 2)))
            if target is not None:
                target = F.pad(target, (0, int((1 + self.size[0] - target.size[1]) / 2)))
        i, j, h, w = self.get_params(img, self.size)

        if target is not None:
            return F.crop(img, i, j, h, w), F.crop(target, i, j, h, w)
        else:
            return F.crop(img, i, j, h, w)

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, padding={1})'.format(self.size, self.padding)


class RandomHorizontalFlip(Transform):
    """Horizontally flip the given PIL Image randomly with a given probability.

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img, target=None):
        """
        Args:
            img (PIL Image): Image to be flipped.
            target (PIL Image): (optional) Target to be flipped

        Returns:
            PIL Image: Randomly flipped image(s).
        """
        if random.random() < self.p:
            if target is not None:
                return F.hflip(img), F.hflip(target)
            else:
                return F.hflip(img)

        if target is not None:
            return img, target
        return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomVerticalFlip(Transform):
    """Vertically flip the given PIL Image randomly with a given probability.

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img, target=None):
        """
        Args:
            img (PIL Image): Image to be flipped.
            target (PIL Image): (optional) Target to be flipped

        Returns:
            PIL Image: Randomly flipped image(s).
        """
        if random.random() < self.p:
            if target is not None:
                return F.vflip(img), F.vflip(target)
            else:
                return F.vflip(img)

        if target is not None:
            return img, target
        return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomResizedCrop(Transform):
    """Crop the given PIL Image to random size and aspect ratio.

    A crop of random size (default: of 0.08 to 1.0) of the original size and a random
    aspect ratio (default: of 3/4 to 4/3) of the original aspect ratio is made. This crop
    is finally resized to given size.
    This is popularly used to train the Inception networks.

    Args:
        size: expected output size of each edge
        scale: range of size of the origin size cropped
        ratio: range of aspect ratio of the origin aspect ratio cropped
        interpolation: Default: PIL.Image.BILINEAR
    """

    def __init__(self, size, scale=(0.08, 1.0), ratio=(3. / 4., 4. / 3.),
                 interpolation=Image.BILINEAR, interpolation_tg = Image.NEAREST):
        self.size = (size, size)
        self.interpolation = interpolation
        self.interpolation_tg = interpolation_tg
        self.scale = scale
        self.ratio = ratio

    @staticmethod
    def get_params(img, scale, ratio):
        """Get parameters for ``crop`` for a random sized crop.
        Args:
            img (PIL Image): Image to be cropped.
            scale (tuple): range of size of the origin size cropped
            ratio (tuple): range of aspect ratio of the origin aspect ratio cropped

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for a random
                sized crop.
        """
        for attempt in range(10):
            area = img.size[0] * img.size[1]
            target_area = random.uniform(*scale) * area
            aspect_ratio = random.uniform(*ratio)

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if random.random() < 0.5:
                w, h = h, w

            if w <= img.size[0] and h <= img.size[1]:
                i = random.randint(0, img.size[1] - h)
                j = random.randint(0, img.size[0] - w)
                return i, j, h, w

        # Fallback
        w = min(img.size[0], img.size[1])
        i = (img.size[1] - w) // 2
        j = (img.size[0] - w) // 2
        return i, j, w, w

    def __call__(self, img, target = None):
        """
        Args:
            img (PIL Image): Image to be cropped and resized.
            target (PIL Image): (optional) Target to be cropped and resized.

        Returns:
            PIL Image: Randomly cropped and resized image(s).
        """
        i, j, h, w = self.get_params(img, self.scale, self.ratio)
        if target is not None:            
            return F.resized_crop(img, i, j, h, w, self.size, self.interpolation), \
                   F.resized_crop(target, i, j, h, w, self.size, self.interpolation_tg)
        return F.resized_crop(img, i, j, h, w, self.size, self.interpolation)

    def __repr__(self):
        interpolate_str = _pil_interpolation_to_str[self.interpolation]
        format_string = self.__class__.__name__ + '(size={0}'.format(self.size)
        format_string += ', scale={0}'.format(tuple(round(s, 4) for s in self.scale))
        format_string += ', ratio={0}'.format(tuple(round(r, 4) for r in self.ratio))
        format_string += ', interpolation={0})'.format(interpolate_str)
        return format_string

class FiveCrop(Transform):
    """Crop the given PIL Image into four corners and the central crop

    .. Note::
         This transform returns a tuple of images and there may be a mismatch in the number of
         inputs and targets your Dataset returns. See below for an example of how to deal with
         this.

    Args:
         size (sequence or int): Desired output size of the crop. If size is an ``int``
            instead of sequence like (h, w), a square crop of size (size, size) is made.

    Example:
         >>> transform = Compose([
         >>>    FiveCrop(size), # this is a list of PIL Images
         >>>    Lambda(lambda crops: torch.stack([ToTensor()(crop) for crop in crops])) # returns a 4D tensor
         >>> ])
         >>> #In your test loop you can do the following:
         >>> input, target = batch # input is a 5d tensor, target is 2d
         >>> bs, ncrops, c, h, w = input.size()
         >>> result = model(input.view(-1, c, h, w)) # fuse batch size and ncrops
         >>> result_avg = result.view(bs, ncrops, -1).mean(1) # avg over crops
    """

    def __init__(self, size):
        self.size = size
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            assert len(size) == 2, "Please provide only two dimensions (h, w) for size."
            self.size = size

    def __call__(self, img, target=None):
        if target is not None:
            return F.five_crop(img, self.size), F.five_crop(target, self.size)
        return F.five_crop(img, self.size)

    def __repr__(self):
        return self.__class__.__name__ + '(size={0})'.format(self.size)


class TenCrop(Transform):
    """Crop the given PIL Image into four corners and the central crop plus the flipped version of
    these (horizontal flipping is used by default)

    .. Note::
         This transform returns a tuple of images and there may be a mismatch in the number of
         inputs and targets your Dataset returns. See below for an example of how to deal with
         this.

    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
        vertical_flip(bool): Use vertical flipping instead of horizontal

    Example:
         >>> transform = Compose([
         >>>    TenCrop(size), # this is a list of PIL Images
         >>>    Lambda(lambda crops: torch.stack([ToTensor()(crop) for crop in crops])) # returns a 4D tensor
         >>> ])
         >>> #In your test loop you can do the following:
         >>> input, target = batch # input is a 5d tensor, target is 2d
         >>> bs, ncrops, c, h, w = input.size()
         >>> result = model(input.view(-1, c, h, w)) # fuse batch size and ncrops
         >>> result_avg = result.view(bs, ncrops, -1).mean(1) # avg over crops
    """

    def __init__(self, size, vertical_flip=False):
        self.size = size
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            assert len(size) == 2, "Please provide only two dimensions (h, w) for size."
            self.size = size
        self.vertical_flip = vertical_flip

    def __call__(self, img, target = None):
        if target is not None:
            return F.ten_crop(img, self.size), F.ten_crop(target, self.size)
        return F.ten_crop(img, self.size, self.vertical_flip)

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, vertical_flip={1})'.format(self.size, self.vertical_flip)


class LinearTransformation(Transform):
    """Transform a tensor image with a square transformation matrix computed
    offline.

    Given transformation_matrix, will flatten the torch.*Tensor, compute the dot
    product with the transformation matrix and reshape the tensor to its
    original shape.

    Applications:
    - whitening: zero-center the data, compute the data covariance matrix
                 [D x D] with np.dot(X.T, X), perform SVD on this matrix and
                 pass it as transformation_matrix.

    Args:
        transformation_matrix (Tensor): tensor [D x D], D = C x H x W
    """

    def __init__(self, transformation_matrix):
        if transformation_matrix.size(0) != transformation_matrix.size(1):
            raise ValueError("transformation_matrix should be square. Got " +
                             "[{} x {}] rectangular matrix.".format(*transformation_matrix.size()))
        self.transformation_matrix = transformation_matrix

    def __call__(self, tensor, target_tensor=None):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be whitened.

        Returns:
            Tensor: Transformed image.
        """
        if target_tensor is not None:
            raise NotImplementedError("LinearTransformation not implemented for tensor pairs.")
        if tensor.size(0) * tensor.size(1) * tensor.size(2) != self.transformation_matrix.size(0):
            raise ValueError("tensor and transformation matrix have incompatible shape." +
                             "[{} x {} x {}] != ".format(*tensor.size()) +
                             "{}".format(self.transformation_matrix.size(0)))
        flat_tensor = tensor.view(1, -1)
        transformed_tensor = torch.mm(flat_tensor, self.transformation_matrix)
        tensor = transformed_tensor.view(tensor.size())
        return tensor

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        format_string += (str(self.transformation_matrix.numpy().tolist()) + ')')
        return format_string


class ColorJitter(Transform):
    """Randomly change the brightness, contrast and saturation of an image.

    Args:
        brightness (float): How much to jitter brightness. brightness_factor
            is chosen uniformly from [max(0, 1 - brightness), 1 + brightness].
        contrast (float): How much to jitter contrast. contrast_factor
            is chosen uniformly from [max(0, 1 - contrast), 1 + contrast].
        saturation (float): How much to jitter saturation. saturation_factor
            is chosen uniformly from [max(0, 1 - saturation), 1 + saturation].
        hue(float): How much to jitter hue. hue_factor is chosen uniformly from
            [-hue, hue]. Should be >=0 and <= 0.5.
    """
    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0):
        self.brightness = brightness
        self.contrast = contrast
        self.saturation = saturation
        self.hue = hue

    @staticmethod
    def get_params(brightness, contrast, saturation, hue):
        """Get a randomized transform to be applied on image.

        Arguments are same as that of __init__.

        Returns:
            Transform which randomly adjusts brightness, contrast and
            saturation in a random order.
        """
        transforms = []
        if brightness > 0:
            brightness_factor = random.uniform(max(0, 1 - brightness), 1 + brightness)
            transforms.append(Lambda(lambda img: F.adjust_brightness(img, brightness_factor)))

        if contrast > 0:
            contrast_factor = random.uniform(max(0, 1 - contrast), 1 + contrast)
            transforms.append(Lambda(lambda img: F.adjust_contrast(img, contrast_factor)))

        if saturation > 0:
            saturation_factor = random.uniform(max(0, 1 - saturation), 1 + saturation)
            transforms.append(Lambda(lambda img: F.adjust_saturation(img, saturation_factor)))

        if hue > 0:
            hue_factor = random.uniform(-hue, hue)
            transforms.append(Lambda(lambda img: F.adjust_hue(img, hue_factor)))

        random.shuffle(transforms)
        transform = Compose(transforms)

        return transform

    def __call__(self, img, target = None):
        """
        Args:
            img (PIL Image): Input image.

        Returns:
            PIL Image: Color jittered image.
        """
        transform = self.get_params(self.brightness, self.contrast,
                                    self.saturation, self.hue)

        if target is not None:
            return transform(img), target
        return transform(img)

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        format_string += 'brightness={0}'.format(self.brightness)
        format_string += ', contrast={0}'.format(self.contrast)
        format_string += ', saturation={0}'.format(self.saturation)
        format_string += ', hue={0})'.format(self.hue)
        return format_string


class RandomRotation(Transform):
    """Rotate the image by angle.

    Args:
        degrees (sequence or float or int): Range of degrees to select from.
            If degrees is a number instead of sequence like (min, max), the range of degrees
            will be (-degrees, +degrees).
        resample ({PIL.Image.NEAREST, PIL.Image.BILINEAR, PIL.Image.BICUBIC}, optional):
            An optional resampling filter.
            See http://pillow.readthedocs.io/en/3.4.x/handbook/concepts.html#filters
            If omitted, or if the image has mode "1" or "P", it is set to PIL.Image.NEAREST.
        expand (bool, optional): Optional expansion flag.
            If true, expands the output to make it large enough to hold the entire rotated image.
            If false or omitted, make the output image the same size as the input image.
            Note that the expand flag assumes rotation around the center and no translation.
        center (2-tuple, optional): Optional center of rotation.
            Origin is the upper left corner.
            Default is the center of the image.
    """

    def __init__(self, degrees, resample=False, resample_tg=False, expand=False, center=None):
        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError("If degrees is a single number, it must be positive.")
            self.degrees = (-degrees, degrees)
        else:
            if len(degrees) != 2:
                raise ValueError("If degrees is a sequence, it must be of len 2.")
            self.degrees = degrees

        self.resample = resample
        self.resample_tg = resample_tg
        self.expand = expand
        self.center = center

    @staticmethod
    def get_params(degrees):
        """Get parameters for ``rotate`` for a random rotation.

        Returns:
            sequence: params to be passed to ``rotate`` for random rotation.
        """
        angle = random.uniform(degrees[0], degrees[1])

        return angle

    def __call__(self, img, target=None):
        """
            img (PIL Image): Image to be rotated.
            target (PIL Image): (optional) Target to be rotated

        Returns:
            PIL Image: Rotated image(s).
        """

        angle = self.get_params(self.degrees)

        if target is not None:
            return F.rotate(img, angle, self.resample, self.expand, self.center), \
                   F.rotate(target, angle, self.resample_tg, self.expand, self.center)
                   # resample = False is by default nearest, appropriate for targets

    def __repr__(self):
        format_string = self.__class__.__name__ + '(degrees={0}'.format(self.degrees)
        format_string += ', resample={0}'.format(self.resample)
        format_string += ', expand={0}'.format(self.expand)
        if self.center is not None:
            format_string += ', center={0}'.format(self.center)
        format_string += ')'
        return format_string


class RandomAffine(Transform):
    """Random affine transformation of the image keeping center invariant

    Args:
        degrees (sequence or float or int): Range of degrees to select from.
            If degrees is a number instead of sequence like (min, max), the range of degrees
            will be (-degrees, +degrees). Set to 0 to desactivate rotations.
        translate (tuple, optional): tuple of maximum absolute fraction for horizontal
            and vertical translations. For example translate=(a, b), then horizontal shift
            is randomly sampled in the range -img_width * a < dx < img_width * a and vertical shift is
            randomly sampled in the range -img_height * b < dy < img_height * b. Will not translate by default.
        scale (tuple, optional): scaling factor interval, e.g (a, b), then scale is
            randomly sampled from the range a <= scale <= b. Will keep original scale by default.
        shear (sequence or float or int, optional): Range of degrees to select from.
            If degrees is a number instead of sequence like (min, max), the range of degrees
            will be (-degrees, +degrees). Will not apply shear by default
        resample ({PIL.Image.NEAREST, PIL.Image.BILINEAR, PIL.Image.BICUBIC}, optional):
            An optional resampling filter.
            See http://pillow.readthedocs.io/en/3.4.x/handbook/concepts.html#filters
            If omitted, or if the image has mode "1" or "P", it is set to PIL.Image.NEAREST.
        fillcolor (int): Optional fill color for the area outside the transform in the output image. (Pillow>=5.0.0)
    """

    def __init__(self, degrees, translate=None, scale=None, shear=None, resample=False, resample_tg=False, fillcolor=0):
        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError("If degrees is a single number, it must be positive.")
            self.degrees = (-degrees, degrees)
        else:
            assert isinstance(degrees, (tuple, list)) and len(degrees) == 2, \
                "degrees should be a list or tuple and it must be of length 2."
            self.degrees = degrees

        if translate is not None:
            assert isinstance(translate, (tuple, list)) and len(translate) == 2, \
                "translate should be a list or tuple and it must be of length 2."
            for t in translate:
                if not (0.0 <= t <= 1.0):
                    raise ValueError("translation values should be between 0 and 1")
        self.translate = translate

        if scale is not None:
            assert isinstance(scale, (tuple, list)) and len(scale) == 2, \
                "scale should be a list or tuple and it must be of length 2."
            for s in scale:
                if s <= 0:
                    raise ValueError("scale values should be positive")
        self.scale = scale

        if shear is not None:
            if isinstance(shear, numbers.Number):
                if shear < 0:
                    raise ValueError("If shear is a single number, it must be positive.")
                self.shear = (-shear, shear)
            else:
                assert isinstance(shear, (tuple, list)) and len(shear) == 2, \
                    "shear should be a list or tuple and it must be of length 2."
                self.shear = shear
        else:
            self.shear = shear

        self.resample = resample
        self.resample_tg = resample_tg
        self.fillcolor = fillcolor

    @staticmethod
    def get_params(degrees, translate, scale_ranges, shears, img_size):
        """Get parameters for affine transformation

        Returns:
            sequence: params to be passed to the affine transformation
        """
        angle = random.uniform(degrees[0], degrees[1])
        if translate is not None:
            max_dx = translate[0] * img_size[0]
            max_dy = translate[1] * img_size[1]
            translations = (np.round(random.uniform(-max_dx, max_dx)),
                            np.round(random.uniform(-max_dy, max_dy)))
        else:
            translations = (0, 0)

        if scale_ranges is not None:
            scale = random.uniform(scale_ranges[0], scale_ranges[1])
        else:
            scale = 1.0

        if shears is not None:
            shear = random.uniform(shears[0], shears[1])
        else:
            shear = 0.0

        return angle, translations, scale, shear

    def __call__(self, img, target=None):
        """
            img (PIL Image): Image to be rotated.
            target (PIL Image): (optional) Target to be rotated

        Returns:
            PIL Image: Rotated image(s).
        """
        ret = self.get_params(self.degrees, self.translate, self.scale, self.shear, img.size)
        if target is not None:
            return F.affine(img, *ret, resample=self.resample, fillcolor=self.fillcolor), \
                   F.affine(target, *ret, resample=self.resample_tg, fillcolor=self.fillcolor)
                   # resample = False is by default nearest, appropriate for targets
        return F.affine(img, *ret, resample=self.resample, fillcolor=self.fillcolor)

    def __repr__(self):
        s = '{name}(degrees={degrees}'
        if self.translate is not None:
            s += ', translate={translate}'
        if self.scale is not None:
            s += ', scale={scale}'
        if self.shear is not None:
            s += ', shear={shear}'
        if self.resample > 0:
            s += ', resample={resample}'
        if self.fillcolor != 0:
            s += ', fillcolor={fillcolor}'
        s += ')'
        d = dict(self.__dict__)
        d['resample'] = _pil_interpolation_to_str[d['resample']]
        return s.format(name=self.__class__.__name__, **d)


class Grayscale(Transform):
    """Convert image to grayscale.

    Args:
        num_output_channels (int): (1 or 3) number of channels desired for output image

    Returns:
        PIL Image: Grayscale version of the input.
        - If num_output_channels == 1 : returned image is single channel
        - If num_output_channels == 3 : returned image is 3 channel with r == g == b

    """

    def __init__(self, num_output_channels=1):
        self.num_output_channels = num_output_channels

    def __call__(self, img, target = None):
        """
        Args:
            img (PIL Image): Image to be converted to grayscale.

        Returns:
            PIL Image: Randomly grayscaled image.
        """
        if target is not None:
            return F.to_grayscale(img, num_output_channels=self.num_output_channels), target
        return F.to_grayscale(img, num_output_channels=self.num_output_channels)

    def __repr__(self):
        return self.__class__.__name__ + '(num_output_channels={0})'.format(self.num_output_channels)


class RandomGrayscale(Transform):
    """Randomly convert image to grayscale with a probability of p (default 0.1).

    Args:
        p (float): probability that image should be converted to grayscale.

    Returns:
        PIL Image: Grayscale version of the input image with probability p and unchanged
        with probability (1-p).
        - If input image is 1 channel: grayscale version is 1 channel
        - If input image is 3 channel: grayscale version is 3 channel with r == g == b

    """

    def __init__(self, p=0.1):
        self.p = p

    def __call__(self, img, target = None):
        """
        Args:
            img (PIL Image): Image to be converted to grayscale.

        Returns:
            PIL Image: Randomly grayscaled image.
        """
        num_output_channels = 1 if img.mode == 'L' else 3
        if random.random() < self.p:
            if target is not None:
                return F.to_grayscale(img, num_output_channels=num_output_channels), target
        if target is not None:
            return img, target
        return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={0})'.format(self.p)




import time as time_pkg
from .torchfun import force_exist
import os


__doc__='utilities for pytorch experiment. Non-mathematical tools.'

def record_experiment(exp_dir='not-specified',record_top_dir='records',logfilename='record.txt',comment=''):
    '''Arguments:
        * exp_dir : directory of this experiment output files, 
                    recorded so that it would be easier for you to find the outcome files 
                    of this experiment later.
        * record_top_dir : create a dir to save all kinds of record logs. default(records). 
                    set to empty string('') to save all record in the current dir.
        * logfilename : filename of this log, usually `train_record` `evaluation_record` etc.
        * comment :string comment added to log paragraph. default is empty string.
    '''
    from sys import argv
    try:
        force_exist(record_top_dir)
        logfilepath = os.path.join(record_top_dir,logfilename)
        if not os.path.exists(logfilepath):
            f = open(logfilepath,'w')
            f.write('科研实验开始::\n')
            f.close()

        with open(logfilepath,'a') as f:
            t = time_pkg.localtime()
            timestr = 'y:%d m:%d d:%d %d:%d:%d' %(t.tm_year,t.tm_mon,t.tm_mday,t.tm_hour,t.tm_min,t.tm_sec)
            command = ' '.join(argv)
            template = \
            '''
-----------------
时间：%s
    命令：%s
    目录名：%s
    效果与结果备注：%s
    原因分析：
    示例：

            ''' % (timestr,command,exp_dir,comment)
            f.write(template)
    except Exception as e:
        print('record_experiment:智能实验助理记录出错了。',e)
    else:
        print('record_experiment:智能实验助理已经帮您记录此次试验。')

def cpu_memory(print_on_screen=True):
    '''total CPU memory usage of current python session.
    returns: (RSS,VMS) In bytes!
            RSS is resident set size, 
            VMS is virtual memory size.
    Notice: 
            return values are in bytes.
            printed values are in MBs.
    '''
    import psutil
    p = psutil.Process(os.getpid())
    rss = p.memory_info().rss
    vms = p.memory_info().vms
    if print_on_screen:
        rssmb = rss/1024/1024
        vmsmb = vms/1024/1024
        print('RAM:%.2fMB Disk:%.2fMB'%(rssmb,vmsmb))
    return rss,vms


def change_suffix(fpath,new_suffix):
    '''
    change the suffix of fpath to new_suffix.
    if the original path fpath has no suffix, then new_suffix will be
    directly applied at the end of the fpath string, joint by separator `.`

    In : change_suffix('xxx/ddd/www.x','aa')
    Out: 'xxx/ddd/www.aa'

    In : change_suffix('xxx/ddd/www','aa')
    Out: 'xxx/ddd/www.aa'
    '''
    fname = os.path.basename(fpath)
    if '.' in fname:
        # xxx/xxx/xxx.suffix
        path_with_name = '.'.join(fpath.split('.')[:-1])
    else:
        path_with_name = fpath
    return path_with_name+'.'+new_suffix

















# LICENSE: MIT
# Author: CHEN Si Yu.
# Date: 2018
# Appended constrains: If the content in this project is used in your source-codes, this author's name must be cited at the begining of your source-code. 

from .torchfun import *
from . import transforms
from . import nn
from . import utils
from . import datasets
from . import functional
from . import ui
from . import path
from .transforms import *
from .nn import *
from .functional import *
from .utils import *
from .ui import *


############ BIND NAMES ##############
list = List
bool = Bool
############ BIND ENDED ##############






import os

name = 'torchfun'

__folder__ = os.path.dirname(__file__)

__ver_fname__ = os.path.join(__folder__,'version')

__version__ = open(__ver_fname__).read()

__all__ = list(locals())
__all__.append('list')

__doc__ = '''`TorchFun` project was initiated long ago and was published in 2018-6-13. 

TorchFun is motivated by the one author loves, who sometimes encounter inconvenience using PyTorch at her work. The author published the project as a python package `TorchFun` on PyPi, so that his beloved could access it whenever and wheresoever needed.

The purposed of TorchFun is to provide functions which are important and convenient, but absent in PyTorch, like some layers and visualization utils.

This project has been undergoing in secret so that the author could credit TorchFun to his beloved as a little supprise of help, one day when this project is well-enriched or gets famous.

Interestingly, The original project name given by the author, is `Torchure`. That is because he was always multi-tasking trivial affairs in school and got scorched, and when he was learning this new framework in hope to help his beloved, he found plenty of issues/missing-functionalities. He felt that this was totally a torture. So, this project was named "Torchure" to satirize the lame PyTorch (you can still found Torchure in PyPi). And, the author hoped, by developing Torchure, his beloved could feel ease even when encountering the crappy-parts of PyTorch.

This history-of-project was appended recently, because his adorable little beloved wants a supprise immediately, or she will keep rolling on the floor.

To c71b05bf46d8772e4488335085a2e7fd.

'''

del_list=[
'os',
'io',
'np',
'torch',
'tqdm',
'division',
'print_function',
'importlib',
'sys']

for del_item in del_list:
    if del_item in __all__:
        __all__.remove(del_item)

del del_list

try:
    __IPYTHON__
    print('Torchfun ',__version__,'.')
    print('Try torchfun.documentation() to get instant help. documentation() accept both torchfun-object and string-query as arguement. and can be invoked without any argument.')
except:
    pass



# LICENSE: MIT
# Author: CHEN Si Yu.
# Date: 2018
# Appended constrains: If the content in this project is used in your source-codes, this author's name must be cited at the begining of your source-code. 

from __future__ import division,print_function
import numpy as np
import torch
import io,os,importlib,sys
from tqdm import tqdm
import types
from .types import argparse_list_type,list_of_int,list_of_float,argparse_bool_type
t = torch

def printf(format_str,*args):
    '''works like C printf.
    '''
    print(format_str % tuple(args))

def print_verbose(*args,verbose=True):
    '''programmable print function. with an option to control
    if the inputs are really printed.
    used to control verbose levels of a function.
    verbose: default True.
    '''
    if verbose:
        print(*args)

def sort_args(args_or_types,types_or_args):
    '''
    This is a very interesting function.
    It is used to support __arbitrary-arguments-ordering__ in TorchFun.

    Input:
        The function takes a list of types, and a list of arguments.

    Returns:
        a list of arguments, with the same order as the types-list.

    Of course, `sort_args` supports arbitrary-arguments-ordering by itself.

    '''
    if (
            (type(args_or_types[0]) is type) # one type
            or 
            ( # type list
                isinstance(args_or_types[0],(list,tuple)) and 
                (type(args_or_types[0][0]) is type)
            )
        ):
        types = args_or_types
        args = types_or_args
    else:
        types = types_or_args
        args = args_or_types

    type_arg_tuples = [(type(a),a) for a in args]
    res=[]
    for t in types:
        found = False
        for type_arg in type_arg_tuples:
            arg_type,arg_val = type_arg
            if issubclass(arg_type,t):
                found=True
                break
        if found:
            res.append(arg_val)
            type_arg_tuples.remove(type_arg)
        else:
            raise TypeError('One of the required argument is of type '+ t.__name__ + ', but none of the given arguments is of this type.')

    return res

def omini_open(path):
    '''
    Opens everything using system default viwer.

    This function can call system GUI to open folders,images,files,videos...
    '''
    import subprocess
    import platform
    if platform.system() == "Windows":
        os.startfile(path)
    elif platform.system() == "Darwin":
        subprocess.Popen(["open", path])
    else:# linux
        if path[:4] == 'http': # browser
            command = 'firefox'
        else:
            command = 'xdg-open'
        subprocess.Popen([command, path])

def to_numpy(tensor,keep_dim=False):
    '''convert a NCHW tensor to NHWC numpy array.
    if the input tensor has only one image,
    then the N dimension will be deleted.

    if the input is not 4-dimensional, 
    the tensor will be simply cat to np array
    '''
    if isinstance(tensor,np.ndarray):
        return tensor 
    elif isinstance(tensor,torch.Tensor):
        tensor = tensor.cpu().detach().numpy()
        dims = len(tensor.shape)
        if dims == 4:
            n,c,h,w = tensor.shape
            tensor = tensor.transpose(0,2,3,1)
            if n==1 and not keep_dim:
                tensor = tensor.squeeze(0)
        elif dims == 3:
            tensor = tensor.transpose(1,2,0)
        return tensor

    else:
        raise Exception('unknown input type:%s'%(tensor.__class__.__name__))

def imread(fname,out_range=(0,1),dtype=torch.float):
    '''read jpg/png/gif/bmp/tiff... image file, and cat to tensor
    function based on imageio.
    args:
        fname: string of the file path
        out_range: tuple, output pixel value range.
        dtype: torch datatype
    Notice:
        the returned image is 1xCxHxW (NCHW).
    '''
    import imageio
    rmin,rmax=out_range
    img = imageio.imread(fname)
    img = (img/255)*(rmax-rmin)+rmin
    img = torch.tensor(img,dtype=dtype)
    if len(img.shape) == 3:
        img.transpose_(2,1)
        img.transpose_(0,1)
    else:
        img.unsqueeze_(0)
    img.unsqueeze_(0)
    return img

def imsave(img_or_dest,dest_or_img):
    '''save torch tensor image(s) or numpy image(s).
    img: can be numpy image, numpy image batch. 
        or single torch-tensor image
        or torch-tensor-image-batch.
        or list/tuple of numpy images
        or list/tuple of pytorch-tensor images

    Notice: images must have non-zero channels, even for grey-scale images (1-channel images).
    
    behaviour: images will be concatenated into a long single image and save into destination file-name.
    
    '''
    img,dest = sort_args((img_or_dest,dest_or_img),((torch.Tensor,np.ndarray,tuple,list),str))

    if isinstance(img,(tuple,list)):
        if len(img)==0:
            raise Exception('No images contained in the input list/tuple')
        if isinstance(img[0],torch.Tensor):
            libcat = torch.cat
            libstack = torch.stack
        elif isinstance(img[0],np.ndarray):
            libcat = np.concatenate
            libstack = np.stack
        else:
            raise Exception('input datatype not understood %s' % str(type(img[0])))
        ## make a batch
        if len(img[0].shape)==4:
            img = libcat(img,0)
        else:
            img = libstack(img)

    if isinstance(img,torch.Tensor):
        dimensions = len(img.shape)
        if dimensions == 4:
            # concatenate into a long image
            img = torch.cat([i for i in img],2)
        dimensions = len(img.shape)
        if dimensions == 3: #make channel last image
            img = img.cpu().numpy().transpose(1,2,0)
    img = _force_image_range(img,out_range=(0,1))
    if img is None:
        return
    import imageio
    imageio.imsave(dest,img)
imwrite = imsave

def _force_image_range(npimg,out_range=(0,1),verbose=True):
    '''input must be numpy image. input tensors will be cat to numpy arrray
    This function automatically detects the domain of the input img batch,
    and force the input to be between the out_range.
    args:
        out_range: (min,max)

    Notice:
        if work under verbose=True, images with suspicious domain will return None
        in order to issue instant debugging.
        if not, the image will be forced into the output range.
        '''
    max_intensity = npimg.max()
    min_intensity = npimg.min()
    if min_intensity>=0 and max_intensity>1 and max_intensity<=255:
        # 0 - 255
        npimg = npimg/255
        print_verbose('TorchFun:imshow:Guessed pixel value range:0~255',verbose=verbose)
    elif min_intensity<0 and min_intensity>=-0.5 and max_intensity>0 and max_intensity <=0.5:
        # -0.5 - 0.5
        npimg += 0.5
        print_verbose('TorchFun:imshow:Guessed pixel value range:-0.5~0.5',verbose=verbose)
    elif min_intensity>=0 and max_intensity<=1:
        print_verbose('TorchFun:imshow:Guessed pixel value range:0~1',verbose=verbose)
    elif min_intensity<-0.5 and min_intensity>=-1 and max_intensity>0.5 and max_intensity <=1:
        # -1 - 1
        npimg = npimg/2
        npimg += 0.5
        print_verbose('TorchFun:imshow:Guessed pixel value range:-1~1',verbose=verbose)
    elif min_intensity>=-1 and max_intensity<=1:
        npimg = npimg/2
        npimg += 0.5
        print_verbose('TorchFun:imshow:Cannot speculate the value-range of this image!, Trying -1~1.',verbose=verbose)
    else:
        msg='TorchFun:imshow:Cannot speculate the value-range of this image. Please normalize the image manually before using imshow.'
        print_verbose(msg,verbose=verbose)
        if verbose:
            return None
        else:# force range to be 0-1
            minv,maxv = npimg.min(),npimg.max()
            npimg = (npimg-minv)/(maxv-minv+1e-11)
    outmin,outmax = out_range
    npimg = npimg*(outmax-outmin) + outmin
    return npimg

def imshow(x,title=None,auto_close=True,cols=8,backend=None):
    '''only deal with torch channel-first image batch,
    title: add title to plot. (Default None)
        title can be string, or any string-able object.
    auto_close: (default True) 
        Close the pyplot session afterwards. 
        Clean the environment just like you had 
        never used matplotlib here.
        if set to False, the plot will remain in the memory for further drawings.
    cols: columns(default 8)
        the width of the output grid, aka, number of images per row.
    backend: None to use default gui. options are:
        WebAgg,  GTK3Agg,
        WX,      GTK3Cairo,
        WXAgg,   MacOSX,
        WXCairo, nbAgg,
        agg,     Qt4Agg,
        cairo,   Qt4Cairo,
        pdf,     Qt5Agg,
        pgf,     Qt5Cairo,
        ps,      TkAgg,
        svg,     TkCairo,
        template        
 
    Usage:
    ```python
        imshow(batch)
        imshow(batch,title=[a,b,c])
        imshow(batch,title='title')
        imshow(batch,auto_close=False) 
    ```
    Warnings:
    ```text
        TorchFun:imshow:Warning, you are using WebAgg backend for Matplotlib. 
        Please consider windowed display SDKs such as TkAgg backend and GTK* backends.
    ```
    This means your matplotlib is using web-browser for figure display. We __strongly__ recommend you to use window-based native display because browser-based backends are fragile and tend to crash. You can change the display mamanger for matplotlib each time you execute your script by:
    ```python
    import matplotlib
    matplotlib.use('TkAgg') # or GTK GTKAgg
    ```
    or permanantly by editing: `site-packages/matplotlib/mpl-data/matplotlibrc` and change backend to `TkAgg`
    
    If your are using Unix-like systems such as MacOSX, you can create ~/.matplotlib/matplotlibrc and add a line: `backend:TkAgg` to it. 

    A full list of available backends can be found at:
    ```python
    import matplotlib
    matplotlib.rcsetup.all_backends
    ```
    and, the TCL/TK GUI library for `tkinter` can be downloaded [here](https://www.tcl.tk/).

    Notice:
        If you use conda to manage your python versions, errors may occur when using TCL/TK.
        That's because conda secretly redirect your global python library path towards its.
        That will cause other stand-alone python versions to search from conda's lib dirs for binaries.
        To solve this, you may have to set:
        
        ```
        export TCL_LIBRARY=/usr/...pythondir.../lib/tcl8.6
        export TK_LIBRARY=/usr/...pythondir.../lib/tcl8.6
        ```
        or on windows:

        ```
        set "TCL_LIBRARY=/usr/...pythondir.../lib/tcl8.6"
        set "TK_LIBRARY=/usr/...pythondir.../lib/tcl8.6"
        ```

        '''
    import torchvision
    if isinstance(x,(list,tuple)):
        x = torch.cat(x,0)
    elif isinstance(x,np.ndarray):
        x = x.transpose((2,0,1))
        x = torch.tensor(x).unsqueeze(0)
    if x.requires_grad:
        x = x.detach()
    x = x.cpu()
    shapes = x.shape
    if len(shapes)==3:
        x = t.unsqueeze(x,dim=0)
    grid = torchvision.utils.make_grid(x,nrow=cols)
    gridnp = grid.numpy()
    gridnp = _force_image_range(gridnp,out_range=(0,1))
    if gridnp is None:
        return

    import matplotlib
    if backend:
        print('using explicitly designated backend:',backend)
        matplotlib.use(backend,force=True)
    else:
        matplotlib.use('TkAgg')
    if matplotlib.get_backend() == 'WebAgg':
        print('TorchFun:imshow:Warning, you are using WebAgg backend for Matplotlib. Please consider windowed display SDKs such as TkAgg backend and GTK* backends.')
    import matplotlib.pyplot as plt

    print("TorchFun:imshow:Notice, To close the window, focus the mouse on the figure and press `q` key.")
    plt.imshow(np.transpose(gridnp,(1,2,0)))
    if title:
        plt.title(title)
    plt.show()
    if auto_close:
        plt.close()
        del plt
    else:
        plt.pause(0.001)


def pil_imshow(arr):
    """
    Simple showing of an image through an external viewer.

    This function is only available if Python Imaging Library (PIL) is installed.

    Uses the image viewer specified by the environment variable
    SCIPY_PIL_IMAGE_VIEWER, or if that is not defined then `see`,
    to view a temporary file generated from array data.

    .. warning::

        This function uses `bytescale` under the hood to rescale images to use
        the full (0, 255) range if ``mode`` is one of ``None, 'L', 'P', 'l'``.
        It will also cast data for 2-D images to ``uint32`` for ``mode=None``
        (which is the default).

    Parameters
    ----------
    arr : ndarray
        Array of image data to show.

    Returns
    -------
    None

    Examples
    --------
    >>> a = np.tile(np.arange(255), (255,1))
    >>> from scipy import misc
    >>> misc.imshow(a)

    Ported and upgraded based on scipy.misc.imshow
    Open-sourced according to the license.

    """
    import PIL
    from PIL import Image
    from PIL.ImageFile import ImageFile
    import tempfile

    if isinstance(arr,torch.Tensor):
        return imshow(arr)

    # to get the tempdir, or the tempdir is None by default.
    fnum,fname = tempfile.mkstemp()
    os.close(fnum)
    os.unlink(fname)

    oldpath = os.path.join(tempfile.tempdir,'torchfun_pil_imshow_tempimg_prefix.png')
    if os.path.exists(oldpath):
        print('cleaning old tmp image.')
        os.unlink(oldpath)

    if isinstance(arr,ImageFile):
        im = arr
    else:
        im = Image.fromarray(arr)
    fnum, fname = tempfile.mkstemp('.png',prefix='torchfun_pil_imshow_tempimg_prefix')
    if im.mode != 'RGB':
        im = im.convert('RGB')

    try:
        im.save(fname)
    except Exception as e:
        print(e)
        raise RuntimeError("Error saving temporary image data.")
    os.close(fnum)
    omini_open(fname)

def imresize(tensor_or_size,size_or_tensor):
    '''stretch pytorch image NCHW, into given shape
    arguments
            tensor: NCHW tensor
            size: tuple or list of [height,width], or a scale factor
    '''
    x,siz = sort_args([tensor_or_size,size_or_tensor],[torch.Tensor,(tuple,list,int,float)])
    if isinstance(siz,(int,float)):
        return torch.nn.functional.interpolate(x,scale_factor=siz,mod='bilinear',align_corners=False)
    else:
        h,w = siz
        return torch.nn.functional.interpolate(x,size=(h,w),mode='bilinear',align_corners=False)

def imcrop_center(tensor_or_size,size_or_tensor):
    '''crop a center patch from pytorch image NCHW
    arguments
            tensor: NCHW tensor
            size: tuple or list of [height,width], or a scale factor
                    given a tuple of height,width, this returns a scaled patch of the 
                    largest center crop.
                    given a factor of scale, this will return a scaled square maximum center crop
    '''
    x,siz = sort_args([tensor_or_size,size_or_tensor],[torch.Tensor,(tuple,list,int,float)])
    *_,h,w = x.shape
    if isinstance(siz,(float,int)):
        out_h = out_w = min(h,w)*siz
        siz = (out_h,out_w)
    else:
        out_h,out_w = siz
    imgratio = h/w
    cropratio = out_h/out_w
    if cropratio>=imgratio:# height filled first
        H = h
        W = int(H / cropratio)
        Hstart = 0
        Wstart = (w - W)//2
    else:
        W = w
        H = W * cropratio
        Wstart = 0
        Hstart = (h-H)//2
    Hend = Hstart+H
    Wend = Wstart+W
    Hstart,Hend,Wstart,Wend = int(Hstart),int(Hend),int(Wstart),int(Wend)
    patch = x[:,:,Hstart:Hend,Wstart:Wend]
    with torch.no_grad():
        return torch.nn.functional.interpolate(patch,size=siz,mode='bilinear',align_corners=False)

def load(a,b):
    '''
    Load weight `a` into model `b`, or load model `b` using weight `a`
    The order of the arguments doesn't matter.
    Example:
        >load('weights.pts',model)
    or
        >load(model,'weights.pts')
    or
        >f = open('weight.pts')
        >load(f,model)
    or
        >load(model,f)

    Return value: None
    Behaviour: the loaded state-dict will be transformed to the same device as model,
            so that torch won't complain about CUDA-memory-insufficient when you just want to load
            weights from disk directly to cpu-model.
    '''
    args = (a,b)
    arg_file_pos = None
    if isinstance(a,torch.nn.Module):
        model = a 
        arg_file_pos = 1
    elif isinstance(b,torch.nn.Module):
        model = b
        arg_file_pos = 0
    else:
        print('TorchFun:load(): Warning! neither of the arguments is pytorch model, abort loading.')
        return
    source = args[arg_file_pos]

    model,source = sort_args((a,b),(torch.nn.Module,(str,io.TextIOWrapper)))
    if isinstance(source,io.TextIOWrapper):
        # file handle input
        source = io.BytesIO(f.read())
    model_device = model.state_dict()[list(model.state_dict().keys())[0]].device
    weights = torch.load(source,map_location=model_device)
    model.load_state_dict(weights)
    return

def save(a,b):
    '''
    save weight `a` into target `b`, or save model `b` into target `a`
    The order of the arguments doesn't matter.
    Example:

        >save('weights.pts',model)

    or

        >save(model,'weights.pts')

    or

        >f = open('weight.pts')
        >save(f,model)

    or

        >save(model,f)

    or

        >save('weights.pts',state_dict)

    or

        >save(state_dict,'weights.pts')

    Return value: None

    '''
    args = (a,b)
    arg_file_pos = None
    if isinstance(a,torch.nn.Module):
        model = a 
        arg_file_pos = 1
    elif isinstance(b,torch.nn.Module):
        model = b
        arg_file_pos = 0
    elif isinstance(a,dict):
        model = a
        arg_file_pos = 1
    elif isinstance(b,dict):
        model = b
        arg_file_pos = 0
    else:
        print('TorchFun:save(): Warning! neither of the arguments is pytorch model, abort loading.')
        return

    target = args[arg_file_pos]
    if isinstance(target,io.TextIOWrapper):
        # file handle input
        target.close()
        target = target.name
    if isinstance(model,dict):
        weights = model
    else:
        weights = model.state_dict()
    torch.save(weights,target)
    return

def count_parameters(model_or_dict_or_param, verbose=True):
    '''Count parameter numer of a module/state_dict/layer/tensor.
    This function can also print the occupied memory of parameters in MBs
    Arguements:
    model_or_dict_or_param: model or state dictionary or model.parameter(), or numpy-array, or tensor.
    Return: parameter amount in python-int
            Returns 0 if datatype not understood
    Usage:
    1. count trainable and untrainbale params
        count_parameters(model)
    same as    
        count_parameters(state_dict)
    2. count only trainable params:
        count_parameters(model.parameters())
    3. count data matrix
        count_parameters(weight_tensor)
        count_parameters(numpy_array)
    Notice:
        return value is parameter Number.

    Alias: parameters()
    '''
    dtype_size={torch.float16:2,torch.float32:4,torch.float64:8,torch.int16:2,torch.int32:4,torch.int64:8,torch.uint8:1}
    if isinstance(model_or_dict_or_param,torch.nn.Module):
        s = model_or_dict_or_param.state_dict()
        w = [s[k] for k in s]
    elif isinstance(model_or_dict_or_param,dict):
        w = [model_or_dict_or_param[i] for i in model_or_dict_or_param]
    elif isinstance(model_or_dict_or_param,types.GeneratorType):
        w = [i for i in model_or_dict_or_param]
    elif getattr(model_or_dict_or_param,'shape',False):
        w= [model_or_dict_or_param.shape]
    else:
        print('TorchFun:count_parameters:Warning, Unknown data type:',model_or_dict.__class__)
        return 0
    num = 0
    KBytes = 0
    for p in w:
        this_num = np.prod(p.shape)
        num += this_num
        dtype = p.dtype if p.dtype in dtype_size else torch.float32
        KBytes += (this_num/1024*dtype_size[dtype])
    if verbose:
        print('Torchfun:parameters:%.2f' % (KBytes/1024),'MBs',num,'params')
    return num

parameters = count_parameters

def hash_parameters(model_or_statdict_or_param,use_sum=False):
    '''return the summary of all variables.
    This is used to detect chaotic changes of weights.
    You can check the sum_parameters before and after some operations, to know
    if there is any change made to the params.

    I use this function to verify gradient behaviours.

    By default, This only hash the trainable parameters!

    arguements:
    module_or_statdict_or_param: torch.nn.module, 
                    or model.state_dict(), 
                    or model.parameters().
    use_sum: return the sum instead of mean value of all params.'''
    m = model_or_statdict_or_param
    means = []
    params = []
    if isinstance(m,torch.nn.Module):
        params = m.parameters()
    elif isinstance(m,dict):
        params = [m[k] for k in m]
    elif isinstance(m,types.GeneratorType):
        params = parameters
    elif isinstance(m,torch.Tensor):
        params = [m]
    else:
        print('TorchFun:hash_parameters:','input type not support:',type(m))

    for p in params:
        if use_sum:
            v = p.sum()
        else:
            v = p.mean()
        means.append(v)

    return float(sum(means))

def vectorize_parameter(model_or_statdict_or_param):
    '''return the vectorized form of all variables.
    This is used to detect chaotic changes of weights.

    arguements:
    module_or_statdict_or_param: torch.nn.module, 
                    or model.state_dict(), 
                    or model.parameters().
   
    '''
    m = model_or_statdict_or_param
    means = []
    params = []
    if isinstance(m,torch.nn.Module):
        params = m.parameters()
    elif isinstance(m,dict):
        params = [m[k] for k in m]
    elif isinstance(m,types.GeneratorType):
        params = parameters
    elif isinstance(m,torch.Tensor):
        params = [m]
    else:
        print('TorchFun:vectorize_parameter:','input type not support:',type(m))
    ps=[]
    for p in params:
        ps.append(p.view(-1,1))

    return torch.cat(ps,0)

def show_layers_parameters(model):
    total_params=0
    print('-----------start-----------')
    for i,name in enumerate(model._modules,1):
        l = model._modules[name]
        l_params = count_parameters(l,verbose=False)
        total_params+=l_params
        print('layer',str(i).ljust(3),l.__class__.__name__.rjust(15),'params:',l_params)
    print('---------------------------')
    print('total parameters:',total_params)
    print('------------end------------')

def show(net,input_shape=(1,3,32,32),logdir='tensorboard',port=8888):
    '''print the network architecture on web-browser, using tensorboardX and tensorboard.
    tensoboard must be install to use this tool.
    this tool will create a noise data according to given input_shape,
    and feed it directly to net, in order to probe its structure.
    network strctures descriptions will be written to logdir.
    a tensorboard daemon will be launched to read the logdir and start a web server
    on given port.
    Notice: 
        input shape must be 
        This program overwrites the system argument lists (sys.argv)
    '''
    try:
        from tensorboard.main import run_main
    except Exception as e:
        raise Exception(str(e)+'\nError importing tensorboard. Maybe your tensorboard is not installed correctly.\
            usually, tensorboard should come with tensorflow. stand-alone tensorboard packages are not stable enough.')
    import tensorboardX as tb
    import sys,shutil,re
    net,input_shape,logdir,port = sort_args([net,input_shape,logdir,port],[torch.nn.Module,(tuple,list),str,int])
    shutil.rmtree(logdir,ignore_errors=True)
    imgs = torch.rand(*input_shape)
    w = tb.SummaryWriter(logdir)
    try:
        w.add_graph(net,imgs)
    except Exception as e:
        raise Exception(str(e)+'\nYour network has problems dealing with input data.\
         It is usually due to wrong input shape or problematic network implementation.\
         Please check your network code for more information.')
    finally:
        w.close()
    args = ['tensorboard','--logdir',logdir,'--port',str(port),'--host','127.0.0.1']
    sys.argv = args
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    print('You may have to delete',logdir,'folder manully.')
    omini_open('http://127.0.0.1:%d'%port)
    try:
        run_main()
    except Exception as e:
        print(e)

class Packsearch(object):
    '''Search names inside a package.
    Given an module object as input:
    > p = Packsearch(torch)
    or
    > p = Packsearch('torch')
    the instance p provide p.search() method. So that you can 
    search everything inside this package
    > p.search('maxpoo')
    or simply
    > p('maxpoo')
    output:
        Packsearch: 35 results found:
        -------------results start-------------
        0        torch.nn.AdaptiveMaxPool1d
        1        torch.nn.AdaptiveMaxPool2d
        2        torch.nn.AdaptiveMaxPool3d
        3        torch.nn.FractionalMaxPool2d
        4        torch.nn.MaxPool1d
        5        torch.nn.MaxPool2d
        ...
    '''
    def __init__(self,module_object,auto_init=True,verbose=False):
        super(self.__class__,self).__init__()
        if isinstance(module_object,types.ModuleType):
            self.root = module_object
        else:
            try:
                self.root = importlib.import_module(module_object)
            except:
                raise TypeError('''Torchfun Packsearch: Wrong Argument!. torchfun.Packsearch is a class. 
                    it accept a torch Module or string as input. Use help(torchfun.Packsearch) to see the usage.
                    Or, you can use torch.packsearch(package,query) function.''')
        self.warned = False # to make warning appear only once.
        self.target = self.root.__name__
        self.name_list = []
        self.name_dict = []
        if auto_init:
            print('Packsearch:','caching names in the module...')
            self.traverse(self.root,search_attributes=verbose)
            self.preprocess_names()
            print('Packsearch:','done, please use search().')

    def preprocess_names(self):
        self.name_list = sorted(self.name_list)
        self.name_dict = [(i.lower(),i) for i in self.name_list]

    def traverse(self,mod,search_attributes=False):
        '''gather all names and store them into a name_list
        search_attributes: whether to include class attributes or method names
        '''
        if type(mod) is not types.ModuleType:
            return
        root_name = mod.__name__+'.' # make `torchvision.` != `torch.`
        root_name_len = len(root_name)
        traversed_mod_names = [root_name]
        stack=[mod]
        while stack:
            m = stack.pop()
            prefix = m.__name__
            if prefix in traversed_mod_names:
                continue

            names = [] 
            try:
                dir_ = dir(m)
                names.extend(dir_)
            except:
                if not self.warned:
                    print('Packsearch:warning:some objects were corrupted by their author, and they are not traversable.')
                    self.warned = True
            names.extend(list(m.__dict__.keys()))
            names = list(set(names))

            for name in names:
                obj = getattr(m,name)
                if type(obj) is types.ModuleType:
                    obj_path = obj.__name__
                    obj_prefix = obj_path[:root_name_len]
                    if obj_prefix == root_name:
                    # only include mods of current package.
                    # packages like io,numpy,os, will be excluded.
                        stack.insert(0,obj) # to be traversed later
                else:
                    # usually do not traverse non-module objects
                    if search_attributes:
                        attrs = dir(obj)
                        items = [prefix+'.'+name+'.'+attr for attr in attrs]
                        self.name_list.extend(items)
                self.name_list.append(prefix+'.'+name)
            del m
            # record that we have visited this module
            traversed_mod_names.append(prefix)

    def dynamic_traverse(self,mod,query,search_attributes=False):
        '''traverse the module and simultaneously search for queried name'''
        if type(mod) is not types.ModuleType:
            return
        root_name = mod.__name__+'.' # make `torchvision.` != `torch.`
        root_name_len = len(root_name)
        traversed_mod_names = [root_name]
        stack=[mod]
        while stack:
            m = stack.pop()
            prefix = m.__name__
            if prefix in traversed_mod_names:
                continue
            names = [] 
            try:
                dir_ = dir(m)
                names.extend[dir_]
            except:
                if not self.warned:
                    print('Packsearch:warning:some objects were corrupted by their author, and they are not traversable.')
                    self.warned = True
            names.extend(list(m.__dict__.keys()))
            names = list(set(names))

            for name in names:
                obj = getattr(m,name)
                if type(obj) is types.ModuleType:
                    obj_path = obj.__name__
                    obj_prefix = obj_path[:root_name_len]
                    if obj_prefix == root_name:
                    # only include mods of current package.
                    # packages like io,numpy,os, will be excluded.
                        stack.insert(0,obj) # to be traversed later
                else:
                    if search_attributes:
                        attrs = dir(obj)
                        items = [prefix+'.'+name+'.'+attr for attr in attrs]
                        for itm in items:
                            if query in itm.lower():
                                yield itm
                        del items
                    del obj
                obj_path = prefix+'.'+name
                if query in obj_path.lower():
                    yield obj_path
            del m
            # record that we have visited this module
            traversed_mod_names.append(prefix)

    def search(self,name):
        q = name.lower()
        res = []
        for k,full_name in tqdm(self.name_dict):
            if q in k:
                res.append(full_name)
        res = sorted(res)
        if not res:
            print('Packsearch:searched',len(self.name_list),'items (instances, functions, modules).',name,'not found.')
        else:
            print('Packsearch:',len(res),'results found:')
            print('-------------results start-------------')
            for i,r in enumerate(res):
                print('%d\t'%i,r)
            print('--------------results end--------------')

    def __len__(self):
        return len(self.name_list)
    def __str__(self):
        return 'Packsearch engine applied on {}, with {} names recoreded. Powered by Torchfun.'.format(self.target,len(self))
    def __repr__(self):
        obj_desc = super(self.__class__,self).__repr__()
        return self.__str__() + os.linesep + 'location:' + obj_desc
    def __call__(self,name):
        return self.search(name)


def packsearch(module_or_str,str_or_module,verbose=False):
    '''Given an module object, and search pattern string as input:
    > packsearch(torch,'maxpoo')
    or
    > packsearch('maxpoo',torch)
    you can search everything inside this package
    output:
        Packsearch: 35 results found:
        -------------results start-------------
        0        torch.nn.AdaptiveMaxPool1d
        1        torch.nn.AdaptiveMaxPool2d
        2        torch.nn.AdaptiveMaxPool3d
        3        torch.nn.FractionalMaxPool2d
        4        torch.nn.MaxPool1d
        5        torch.nn.MaxPool2d
        ...
    '''
    mod,name = sort_args([types.ModuleType,str],[module_or_str,str_or_module])
    p = Packsearch(mod,auto_init=False)
    res_counter = 0
    print('Packsearch: dynamic searching start ...')
    for r in p.dynamic_traverse(mod,name.lower(),search_attributes=verbose):
        print(r)
        res_counter += 1
    if res_counter == 0:
        print('Packsearch: no result found')
    else:
        print('Packsearch: search done,',res_counter,'results found.')

def force_exist(dirname,verbose=True):
    '''force a directory to exist.
    force_exist can automatically create directory with any depth.
    Arguements:
        dirname: path of the desired directory
        verbose: print every directory creation. default True.
    Usage:
        force_exist('a/b/c/d/e/f')
        force_exist('a/b/c/d/e/f',verbose=False)
    '''
    
    if dirname == '' or dirname == '.':
        return True
    top = os.path.dirname(dirname)
    force_exist(top)
    if not os.path.exists(dirname):
        if verbose:
            print('creating',dirname)
        os.makedirs(dirname)
        return False
    else:
        return True

def whereis(module_or_string,open_gui=True):
    '''
    find the source file location of a module
    arguments:
        module_or_string: target module object, or it's string path like `torch.nn`
        open_gui: open the folder with default window-manager.
    returns:
        module file name, or None

    '''
    mors = module_or_string
    mod = None
    if isinstance(module_or_string,types.ModuleType):
        mod = module_or_string
    elif ((not isinstance(module_or_string,str)) and
     hasattr(module_or_string,'__module__')):
        module_or_string = module_or_string.__module__
    
    if isinstance(module_or_string,str):
        try:
            mod = importlib.import_module(module_or_string)
        except:
            s = '.'.join(module_or_string.split('.')[:-1])
            print('TorchFun:warning:',module_or_string,'is not a module. Maybe it is a node class. TorchFun is trying to parse its father prefix:',s)
            try:
                mod = importlib.import_module(s)
            except Exception as e:
                print('TorchFun:error:','father prefix',s,'is not valid')
                raise e
         
    if not hasattr(mod,'__file__'):
        print('TorchFun:error:invalid arguement',mors)
        return None

    fname = mod.__file__
    dirname = os.path.dirname(mod.__file__)
    print(fname)
    if open_gui:
        omini_open(dirname)

    return fname
    
def tf_session(allow_growth=True):
    '''Used to create tensorflow session that does not stupidly and unresonably consume all gpu-memeory.
    returns:
        a tensorflow session consuming dynamic gpu memory.'''
    try:
        tensorflow = importlib.import_module('tensorflow')
    except Exception as e:
        print('tensorflow not installed, cannot provide tensorflow session instance.')
        raise e
    config = tensorflow.ConfigProto()
    config.gpu_options.allow_growth = allow_growth
    return tensorflow.Session(config=config)

class Options(object):
    '''A simple yet effective option class for debugging use.
    key features: you can set attributes to it directly.
    like:
            o = Options()
            o.what=1
            o.hahah=123

    '''
    def __init__(self):
        super().__init__()
    def __setattr__(self,name,value):
        if hasattr(self,name):
            super().__setattr__(name,value)
        else:
            self.__dict__[name]=value
    def __str__(self):
        return str(self.__dict__)
    def __repr__(self):
        contents=['\t%s:%s'%(name,str(self.__dict__[name])) for name in self.__dict__]
        contents.insert(0,'Options containing:')
        return os.linesep.join(contents)

def options(*args,**kws):
    '''warpping class for Options. this function returns an option object with attributes
    set according to the input key-value arguments. 
    please refer to Option class for more information'''
    if len(args)!=0:
        print('warning:options must be keywords pairs, not anoymous value list')
    o = Options()
    o.__dict__.update(kws)
    return o

class Unimodel(torch.nn.Module):
    '''this class is used to gather your multiple models, so that
    you can save/load them together.

    usage:

        unimodel = Unimodel(resnet1,resnet2,resnet3)
        unimodel.save('xxxx.pth')
        unimodel.load('ssss.pth')

    '''
    def __init__(self,*models,**named_models):
        super().__init__()
        for i,m in enumerate(models):
            self.__setattr__('model%d'%i,m)
        for name in named_models:
            self.__setattr__(name,named_models[name])
    def save(self,path):
        '''same as torchfun save()'''
        save(self,path)
    def load(self,path):
        '''same as torchfun load()'''
        load(self,path)


















# try not to override default bool type
# TODO: change bool, or somehow it will 
# overwrite all occurrence of bool above
from .types import Bool,TorchEasyList
List = TorchEasyList

# try not to override default list type
##################### Last function ####################


def documentation(search=None):
    ''' help documentation on Torchfun
    Argument:
        search: give None to go to the latest doc site
                give string or object to search the object
    '''
    index_page = 'https://sorenchiron.github.io/torchfun/genindex.html'
    search_page = 'https://sorenchiron.github.io/torchfun/search.html?q='
    if search is None:
        omini_open(index_page)
    else:
        if not isinstance(search,str):
            search = search.__name__
            search_url = search_page+search
            omini_open(search_url)



import torch
import torchfun as tf 
from torchfun import *
def max_min_norm(x,to_max,to_min,eps=1e-5):
    '''scale the input x into range [to_min,to_max].
    Arguments:
        to_max: target expect max value of the output
        to_min: target expect min value of the output
    '''
    b,c,h,w = x.shape
    x = flatten(x)
    xmax,_ = x.max(1)
    xmin,_ = x.min(1)
    x = x.transpose(0,1)
    x = (x-xmin)/(xmax-xmin+eps)
    x = x * (to_max - to_min) + to_min
    x = x.transpose(0,1)
    return x.view(b,c,h,w)

class MaxMinNorm(Clip):
    __doc__ = max_min_norm.__doc__
    def __init__(self,max_or_min,min_or_max,eps=1e-5):
        super(MaxMinNorm,self).__init__(max_or_min,min_or_max)
        eps = torch.tensor(eps,dtype=torch.float)
        self.register_buffer(name='eps',tensor=eps)
    def forward(self,x):
        return max_min_norm(x,self.maxv,self.minv,self.eps)


x = torch.rand(1,2,3,4)
x0 = x[0]
x0min = x0.min()
x0max = x0.max()
y0 = ((x0-x0min)/(x0max-x0min+1e-5))*(1 - -1) + -1
print(tf.hash_parameters(y0))

m = MaxMinNorm(-1,1)
out = m(x)

print(tf.hash_parameters(out[0]))

print((out[0]-y0).abs().sum())


# thanks to https://github.com/Po-Hsun-Su/pytorch-ssim
import torch
import torch.nn.functional as F
import numpy as np
from math import exp
from .torchfun import _force_image_range

def gaussian(window_size, sigma):
    gauss = torch.Tensor([exp(-(x - window_size//2)**2/float(2*sigma**2)) for x in range(window_size)])
    return gauss/gauss.sum()

def create_window(window_size, channel):
    _1D_window = gaussian(window_size, 1.5).unsqueeze(1)
    _2D_window = _1D_window.mm(_1D_window.t()).float().unsqueeze(0).unsqueeze(0)
    window = _2D_window.expand(channel, 1, window_size, window_size).contiguous()
    return window

def _ssim(img1, img2, window, window_size, channel, size_average = True):
    window = window.to(img1.device)
    mu1 = F.conv2d(img1, window, padding = window_size//2, groups = channel)
    mu2 = F.conv2d(img2, window, padding = window_size//2, groups = channel)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1*mu2

    sigma1_sq = F.conv2d(img1*img1, window, padding = window_size//2, groups = channel) - mu1_sq
    sigma2_sq = F.conv2d(img2*img2, window, padding = window_size//2, groups = channel) - mu2_sq
    sigma12 = F.conv2d(img1*img2, window, padding = window_size//2, groups = channel) - mu1_mu2

    C1 = 0.01**2
    C2 = 0.03**2

    ssim_map = ((2*mu1_mu2 + C1)*(2*sigma12 + C2))/((mu1_sq + mu2_sq + C1)*(sigma1_sq + sigma2_sq + C2))

    if size_average:
        return ssim_map.mean()
    else:
        return ssim_map.mean(1).mean(1).mean(1)

class SSIM(torch.nn.Module):
    '''structural similarity'''
    def __init__(self, window_size = 11, size_average = True):
        super(SSIM, self).__init__()
        self.window_size = window_size
        self.size_average = size_average
        self.channel = 1
        self.window = create_window(window_size, self.channel)

    def forward(self, img1, img2):
        (_, channel, _, _) = img1.size()

        if channel == self.channel and self.window.data.type() == img1.data.type():
            window = self.window
        else:
            window = create_window(self.window_size, channel)
            
            if img1.is_cuda:
                window = window.cuda(img1.get_device())
            window = window.type_as(img1)
            
            self.window = window
            self.channel = channel


        return _ssim(img1, img2, window, self.window_size, channel, self.size_average)

def ssim(img1, img2, window_size = 11, size_average = True):
    (_, channel, _, _) = img1.size()
    window = create_window(window_size, channel)
    
    if img1.is_cuda:
        window = window.cuda(img1.get_device())
    window = window.type_as(img1)
    
    return _ssim(img1, img2, window, window_size, channel, size_average)

# thanks to https://github.com/lizhengwei1992/MS_SSIM_pytorch/blob/master/loss.py
class MS_SSIM(torch.nn.Module):
    def __init__(self, size_average = True, max_val = 255):
        super(MS_SSIM, self).__init__()
        self.size_average = size_average
        self.channel = 3
        self.max_val = max_val
    def _ssim(self, img1, img2, size_average = True):

        _, c, w, h = img1.size()
        window_size = min(w, h, 11)
        sigma = 1.5 * window_size / 11
        window = create_window(window_size, sigma, self.channel).cuda()
        mu1 = F.conv2d(img1, window, padding = window_size//2, groups = self.channel)
        mu2 = F.conv2d(img2, window, padding = window_size//2, groups = self.channel)

        mu1_sq = mu1.pow(2)
        mu2_sq = mu2.pow(2)
        mu1_mu2 = mu1*mu2

        sigma1_sq = F.conv2d(img1*img1, window, padding = window_size//2, groups = self.channel) - mu1_sq
        sigma2_sq = F.conv2d(img2*img2, window, padding = window_size//2, groups = self.channel) - mu2_sq
        sigma12 = F.conv2d(img1*img2, window, padding = window_size//2, groups = self.channel) - mu1_mu2

        C1 = (0.01*self.max_val)**2
        C2 = (0.03*self.max_val)**2
        V1 = 2.0 * sigma12 + C2
        V2 = sigma1_sq + sigma2_sq + C2
        ssim_map = ((2*mu1_mu2 + C1)*V1)/((mu1_sq + mu2_sq + C1)*V2)
        mcs_map = V1 / V2
        if size_average:
            return ssim_map.mean(), mcs_map.mean()

    def ms_ssim(self, img1, img2, levels=5):

        weight = torch.Tensor([0.0448, 0.2856, 0.3001, 0.2363, 0.1333]).type_as(img1).to(img1.device)

        msssim = torch.Tensor(levels,).type_as(img1).to(img1.device)
        mcs = torch.Tensor(levels,).type_as(img1).to(img1.device)
        for i in range(levels):
            ssim_map, mcs_map = self._ssim(img1, img2)
            msssim[i] = ssim_map
            mcs[i] = mcs_map
            filtered_im1 = F.avg_pool2d(img1, kernel_size=2, stride=2)
            filtered_im2 = F.avg_pool2d(img2, kernel_size=2, stride=2)
            img1 = filtered_im1
            img2 = filtered_im2

        value = (torch.prod(mcs[0:levels-1]**weight[0:levels-1])*
                                    (msssim[levels-1]**weight[levels-1]))
        return value


    def forward(self, img1, img2):

        return self.ms_ssim(img1, img2)

def psnr(x,y,eps=1e-11):
    x = _force_image_range(x,(0,255))
    y = _force_image_range(y,(0,255))

    mse = torch.nn.functional.mse_loss(x,y)

    vmax_power2 = 65025.0 #2^bits-1,  255.0^2 = 

    return 10*torch.log10(vmax_power2/(mse+eps))




import os

def get_dir_fname_suffix(path):
    ''''''
    dirname = os.path.dirname(path)
    basename = os.path.basename(path)
    if not '.' in basename:
        fname = basename
        suffix = None
    else:
        suffix = basename.split('.')[-1]
        fname = basename[:-len(suffix)-1]
    return dirname,fname,suffix

def add_tag(path,tag):
    '''add tag to the end of the filename, before the suffix(file type mark)  
    dir_path/filename{tag}.suffix'''
    dirname,fname,suffix = get_dir_fname_suffix(path)
    if suffix is not None:
        fname = fname+'tag.'+suffix
    else:
        fname = fname+'tag'
    return os.path.join(dirname,fname)

def add_prefix(path,prefix):
    ''''''
    dirname,fname,suffix = get_dir_fname_suffix(path)
    if suffix is not None:
        fname = prefix+fname+'.'+suffix
    else:
        fname = prefix+fname
    return os.path.join(dirname,fname)

def change_suffix(path,suffix):
    ''''''
    dirname,fname,_ = get_dir_fname_suffix(path)
    return os.path.join(dirname,fname+'.'+suffix)



import setuptools
import pickle
import os


def increase_version_number():
    if os.path.exists('version'):
        version = open('version','r')
    else:
        version = '0.0.0'
    print('previous version',version)
    release,updates,fixes = version.read().split('.')
    updates_increase=release_increase=0
    if '+' in fixes:
        fixes = '-1'
        updates_increase = 1
    if '+' in updates:
        updates = '0'
        fixes = '-1'
        release_increase = 1

    fixes = str(int(fixes)+1)
    updates = str(int(updates)+updates_increase)
    release = str(int(release)+release_increase)

    version.close()
    new_version = open('version','w')
    version = '.'.join([release,updates,fixes])
    new_version.write(version)
    print('new version',version)
    return version

def write_version(v):
    with open('torchfun/version','w') as f:
        f.write(v)

def write_install(v):
    fname = 'torchfun-%s-py3-none-any.whl' % (v)
    fpath = os.path.join('dist',fname)
    with open('local_install.bat','w') as f:
        f.write('pip install '+fpath)

version = increase_version_number()
write_version(version)
write_install(version)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="torchfun",
    version=version,
    author="CHEN Si Yu",
    author_email="sychen@zju.edu.cn",
    description="A collection of small functions that supplements torch functionality",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sorenchiron/torchfun",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={
                '':['version']
                },
    classifiers=(
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License"
    ),
    install_requires=[
        'numpy',
        'torch',
        'torchvision',
        'matplotlib>=3.0.0',
        'Pillow',
        'scipy',
        'tqdm',
        'psutil',
        'tensorboardX<=1.5',
        'imageio',
        'img2pdf'],
    entry_points={
        'console_scripts': [
            'imgformat = torchfun.tools.imgformat:main'
        ],
        'gui_scripts': [
        ]
    }
)


# LICENSE: MIT
# Author: CHEN Si Yu.
# Date: 2018
# Appended constrains: If the content in this project is used in your source-codes, this author's name must be cited at the begining of your source-code. 

from __future__ import division,print_function
import torch
from . import functional
from .functional import *
from .torchfun import sort_args
from time import time
from collections import OrderedDict
import itertools

__doc__ = '''Neural Network related layers/functions/classes
that are compatible with all pyTorch usage.
'''

class DebugAgent(torch.nn.Module):
    '''wrapper around layers.
    this wrapper hooks the forward function, to measure time 
    consumption of this layer.'''
    def __init__(self,obj):
        super(DebugAgent,self).__init__()
        self.obj = obj
        self.name = obj.__class__.__name__
        self.bind(obj)
    def bind(self,obj):
        self.__dict__.update(obj.__dict__)
        self.deep_names = {name:getattr(obj,name) for name in obj.__dir__()}
        self.__dict__.update(self.deep_names)
    def __call__(self,*argv,**kw):
        start = time()
        ret = self.deep_names['forward'](*argv,**kw)
        end = time()
        print('debug:',self.name.rjust(15)+'\t:%.5f sec'%(end-start))
        return ret
    def __getattr__(self,name):
        if name in self.__dict__:
            return self.__dict__[name]
        elif name in self.deep_names:
            return self.deep_names[name]
        else:
            return getattr(self.obj,name)


class Module(torch.nn.Module):
    '''More debugging/controlling methods with complete original
    features from torch.nn.Module
    provides:
        * debug mode: inspect running time layer by layer.
        * release mode: back to normal from debug mode.
        * freeze layers: set layers to be in-trainable
        * unfreeze: as the name implies.
        * unfreeze_all: so as the name implies.

    Notice: you can safely change the base class from torchfun module
            back to torch module, when you want to publish the model.
            the state_dict will be loaded correctly and the forward() will
            function the same.
    Hint:   Consider establishing a BaseClass global variable at the top of your 
            code. Use a argument parser to select between torchfun.nn.Module and torch.nn.Module,
            so that the following classes follows the specified base class

            Example::
                    | import torch.nn.Module as ReleaseModule
                    | import torchfun.nn.Module as DevModule    
                    | from sys import argv
                    | if argv[1] == 'develop':
                    |     Base = DevModule
                    | elif argv[1] == 'release':
                    |     Base = ReleaseModule
                    | 
                    | class MyModel(Base):
                    |     ...

    '''
    def __init__(self):
        super(Module,self).__init__()
        self._debug=False
        self._frozen_names=[]
        if hasattr(self,'forward') and self.forward.__doc__ == None:
            self.forward.__doc__ = self.__class__.__doc__

    def debug(self):
        '''turn on debug mode.
        allow detailed timing report of forward()'''
        if self._debug:
            return
        else:
            self._debug=True
            self._modules_bak=self._modules
            self.debug_modules = OrderedDict()
            for name in self._modules:
                self.debug_modules[name]=DebugAgent(self._modules[name])
            self._modules = self.debug_modules
            self._update_using_modules_dict(self.debug_modules)
    def release(self):
        if self._debug:
            del self._modules
            self._modules = self._modules_bak
            self._update_using_modules_dict(self._modules_bak)
            self._debug = False
    def _update_using_modules_dict(self,odict):
        for name in odict:
            self.__setattr__(name,odict[name])
    def freeze(self,*obj_or_name):
        for o in obj_or_name:
            self._freeze(o,True)
    def unfreeze(self,*obj_or_name):
        for o in obj_or_name:
            self._freeze(o,False)
    def unfreeze_all(self):
        self._frozen_names = []
    def _freeze(self,obj_or_name,freeze=True):
        modules = {self._modules:name for name in self._modules}
        if obj_or_name in self._modules:
            name = obj_or_name
        elif obj_or_name in modules:
            name = modules[obj_or_name]
        else:
            raise Exception('layer or name:%s not in this module'%obj_or_name)
        if freeze:
            if name not in self._frozen_names:
                self._frozen_names.append(name)
        else:
            if name in self._frozen_names:
                del self._frozen_names[name]
        del modules
    
    def train(self,mode=True):
        super(Module,self).train(mode)
        for name in self._frozen_names:
            getattr(self,name).eval()

    def parameters(self):
        if self._frozen_names:
            generators = [self._modules[name].parameters() for name in self._modules if name not in self._frozen_names]
            return itertools.chain(*generators)
        else:
            return super(Module,self).parameters()


class Flatten(torch.nn.Module):
    '''Flatten module
    Usage:
        flat = Flatten()
        out = flat(x)
    '''
    def __init__(self):
        super(Flatten,self).__init__()

    def forward(self,x):
        return flatten(x)

class Subpixel(torch.nn.Module):
    '''Unfold channel/depth dimensions to enlarge the feature map
    Notice: Output size is deducted. 
    The size of the unfold square is automatically determined
    e.g. :
        images: 100x16x16x9.  9=3x3 square
        subpixel-out: 100x48x48x1
    Arguement:
        out_channels, channel number of output feature map
        stride: enlarging ratio of spatial dimensions. stride=2 outputs x4 img area. If provided, out_channels will be ignored.'''
    def __init__(self,out_channels=1,stride=None):
        super(Subpixel,self).__init__()
        self.out_channels = int(out_channels)
        self.stride = int(stride)
        if stride:
            self.patch_pixels = int(stride**2)
        else:
            self.patch_pixels = stride
    def forward(self,x):
        if not self.stride:
            return subpixel(x,out_channels=self.out_channels)
        else:
            c = int(x.size(1)) // self.patch_pixels
            return subpixel(x,out_channels=c)

class Conv2dDepthShared(torch.nn.Conv2d):
    r"""
    Applies a 2D convolution over an input signal composed of several input
    planes.

    Share the kernel along depth/channel direction.
    Conv2dDepthShared divides input images into multiple sub-layers(trunks) along depth axis, and use shared kernel to process each depth trunk.
    
    In the simplest case, the output value of the layer with input size
    :math:`(N, C_{in}, H, W)` and output :math:`(N, C_{out}, H_{out}, W_{out})`
    can be precisely described as:

    .. math::

        \begin{equation*}
        \text{out}(N_i, C_{out_j}) = \text{bias}(C_{out_j}) +
                                \sum_{k = 0}^{C_{in} - 1} \text{weight}(C_{out_j}, k) \star \text{input}(N_i, k)
        \end{equation*},

    where :math:`\star` is the valid 2D `cross-correlation`_ operator,
    :math:`N` is a batch size, :math:`C` denotes a number of channels,
    :math:`H` is a height of input planes in pixels, and :math:`W` is
    width in pixels. 
    The `weight` and `bias` matrix of a Conv2dDepthShared are low-rank. 
    They share trunks of digits repeatitively inside their matrices. 

    Example:
        Conv(in=3,out=9,k=3,s=1) will create kernel weight with size of (9x3x5x5)
        kernel of a Depth-shared-Conv2d only has 3x1x5x5 parameters.

    * :attr:`stride` controls the stride for the cross-correlation, a single
      number or a tuple.

    * :attr:`padding` controls the amount of implicit zero-paddings on both
      sides for :attr:`padding` number of points for each dimension.

    * :attr:`dilation` controls the spacing between the kernel points; also
      known as the à trous algorithm. It is harder to describe, but this `link`_
      has a nice visualization of what :attr:`dilation` does.

    * :attr:`groups` controls the connections between inputs and outputs.
      :attr:`in_channels` and :attr:`out_channels` must both be divisible by
      :attr:`groups`. For example,

        * At groups=1, all inputs are convolved to all outputs.
        * At groups=2, the operation becomes equivalent to having two conv
          layers side by side, each seeing half the input channels,
          and producing half the output channels, and both subsequently
          concatenated.
        * At groups= :attr:`in_channels`, each input channel is convolved with
          its own set of filters (of size
          :math:`\left\lfloor\frac{\text{out_channels}}{\text{in_channels}}\right\rfloor`).

    The parameters :attr:`kernel_size`, :attr:`stride`, :attr:`padding`, :attr:`dilation` can either be:

        - a single ``int`` -- in which case the same value is used for the height and width dimension
        - a ``tuple`` of two ints -- in which case, the first `int` is used for the height dimension,
          and the second `int` for the width dimension

    .. note::

         Depending of the size of your kernel, several (of the last)
         columns of the input might be lost, because it is a valid `cross-correlation`_,
         and not a full `cross-correlation`_.
         It is up to the user to add proper padding.

    .. note::

         The configuration when `groups == in_channels` and `out_channels == K * in_channels`
         where `K` is a positive integer is termed in literature as depthwise convolution.

         In other words, for an input of size :math:`(N, C_{in}, H_{in}, W_{in})`, if you want a
         depthwise convolution with a depthwise multiplier `K`,
         then you use the constructor arguments
         :math:`(\text{in_channels}=C_{in}, \text{out_channels}=C_{in} * K, ..., \text{groups}=C_{in})`

    Args:
        in_channels (int): Number of channels in the input image, inchannels must can be divided by trunks.
        out_channels (int): Number of channels produced by the convolution. out channels must can be divided by trunks.
        trunks (int): Number of trunks a single image is divided into (along depth). All trunks inside an image share same weight/bias.
        kernel_size (int or tuple): Size of the convolving kernel
        stride (int or tuple, optional): Stride of the convolution. Default: 1
        padding (int or tuple, optional): Zero-padding added to both sides of the input. Default: 0
        dilation (int or tuple, optional): Spacing between kernel elements. Default: 1
        groups (int, optional): Number of blocked connections from input channels to output channels. Default: 1
        bias (bool, optional): If ``True``, adds a learnable bias to the output. Default: ``True``

    Shape:
        - Input: :math:`(N, C_{in}, H_{in}, W_{in})`
        - Output: :math:`(N, C_{out}, H_{out}, W_{out})` where

          .. math::
              H_{out} = \left\lfloor\frac{H_{in}  + 2 * \text{padding}[0] - \text{dilation}[0]
                        * (\text{kernel_size}[0] - 1) - 1}{\text{stride}[0]} + 1\right\rfloor

              W_{out} = \left\lfloor\frac{W_{in}  + 2 * \text{padding}[1] - \text{dilation}[1]
                        * (\text{kernel_size}[1] - 1) - 1}{\text{stride}[1]} + 1\right\rfloor

    Attributes:
        weight (Tensor): the learnable weights of the module of shape
                         (out_channels, in_channels, kernel_size[0], kernel_size[1])
        bias (Tensor):   the learnable bias of the module of shape (out_channels)

    Examples::

        >>> # With square kernels and equal stride
        >>> m = nn.Conv2d(16, 33, 3, stride=2)
        >>> # non-square kernels and unequal stride and with padding
        >>> m = nn.Conv2d(16, 33, (3, 5), stride=(2, 1), padding=(4, 2))
        >>> # non-square kernels and unequal stride and with padding and dilation
        >>> m = nn.Conv2d(16, 33, (3, 5), stride=(2, 1), padding=(4, 2), dilation=(3, 1))
        >>> input = torch.randn(20, 16, 50, 100)
        >>> output = m(input)

    .. _cross-correlation:
        https://en.wikipedia.org/wiki/Cross-correlation

    .. _link:
        https://github.com/vdumoulin/conv_arithmetic/blob/master/README.md
    
    """

    def __init__(self, in_channels, out_channels, trunks, kernel_size, 
        stride=1,padding=0, 
        dilation=1, groups=1, bias=True):
        if out_channels % trunks >0:
            message = self.__class__+':'+'out channels should be dividable by trunks!'
            print(message)
            raise Exception(message)
        elif in_channels % trunks >0:
            message = self.__class__+':'+'in channels should be dividable by trunks!'
            print(message)
            raise Exception(message)
        slice_depth = in_channels // trunks
        slice_out = out_channels // trunks

        super(Conv2dDepthShared, self).__init__(slice_depth, slice_out, 
            kernel_size, 
            stride,padding, 
            dilation=dilation, groups=groups, bias=bias)

        self.trunks = trunks
        self.slice_depth = slice_depth
        self.slice_out = slice_out
        self.conv2d_fn = torch.nn.functional.conv2d
        self.in_channels = int(slice_depth * trunks)
        self.out_channels = int(slice_out * trunks)

    def forward(self, x):
        b,c,h,w = x.shape
        xslices = x.contiguous().view(b*self.trunks,
                                    self.slice_depth,h,w)
        yraw = self.conv2d_fn(xslices, self.weight, 
                            self.bias, self.stride,
                            self.padding, self.dilation, self.groups)
        _,_,h,w = yraw.shape
        y = yraw.contiguous().view(b,self.out_channels,h,w)
        return y

class Squeeze(torch.nn.Module):
    '''squeeze(dim=None) -> Tensor

    Returns a tensor with all the dimensions of :attr:`input` of size `1` removed.

    For example, if `input` is of shape:
    :math:`(A \times 1 \times B \times C \times 1 \times D)` then the `out` tensor
    will be of shape: :math:`(A \times B \times C \times D)`.

    When :attr:`dim` is given, a squeeze operation is done only in the given
    dimension. If `input` is of shape: :math:`(A \times 1 \times B)`,
    `squeeze(input, 0)` leaves the tensor unchanged, but :func:`squeeze(input, 1)` will
    squeeze the tensor to the shape :math:`(A \times B)`.

    .. note:: As an exception to the above, a 1-dimensional tensor of size 1 will
              not have its dimensions changed.

    .. note:: The returned tensor shares the storage with the input tensor,
          so changing the contents of one will change the contents of the other.
    
    Example::

        >>> x = torch.zeros(2, 1, 2, 1, 2)
        >>> x.size()
        torch.Size([2, 1, 2, 1, 2])
        >>> y = torch.squeeze(x)
        >>> y.size()
        torch.Size([2, 2, 2])
        >>> y = torch.squeeze(x, 0)
        >>> y.size()
        torch.Size([2, 1, 2, 1, 2])
        >>> y = torch.squeeze(x, 1)
        >>> y.size()
        torch.Size([2, 2, 1, 2]) 

    '''
    def __init__(self,dim=None):
        super(Squeeze,self).__init__()
        self.dim = dim
    def forward(self,x):
        if self.dim:
            return x.squeeze_(self.dim)
        else:
            return x.squeeze_()

#class AbsMax(torch.nn.Module):
#    '''
#    TODO: not fully implemented
#    '''
#    def __init__(self,dim=1):
#        super(AbsMax,self).__init__()
#        self.dim = dim
#    def forward(self,x):
#        signs = x.sign()
#        absx = x.abs()

class Clip(torch.nn.Module):
    __doc__=clip_.__doc__
    def __init__(self,max_or_min,min_or_max,dtype=torch.float,keep_grad=True):
        super(Clip,self).__init__()
        minv,maxv = sorted([max_or_min,min_or_max])
        minv,maxv = torch.tensor([minv,maxv],dtype=dtype)
        self.register_buffer(name='minv',tensor=minv)
        self.register_buffer(name='maxv',tensor=maxv)
        if keep_grad:
            self.forward = self.forward_keep_grad
        else:
            self.forward = self.forward_clip_grad

    def forward_clip_grad(self,x):
        return torch.min(torch.max(x,self.minv),self.maxv)
    def forward_keep_grad(self,x):
        minv,maxv = self.minv,self.maxv
        in_tensor = x
        greater_pos = (in_tensor.detach()>maxv).type_as(in_tensor)
        diff = (maxv-in_tensor.detach())*greater_pos
        in_tensor = in_tensor+diff

        less_pos = (in_tensor.detach()<minv).type_as(in_tensor)
        diff = (minv-in_tensor.detach())*less_pos
        in_tensor = in_tensor+diff
        return in_tensor

class NO_OP(torch.nn.Module):
    '''A Module that repersents NO-Operation NO-OP.
    NO-OP is needed when programmers want customizable dynamic assembling
    of models. 
    To disable some layers, instead of using multiple `if` clauses, nn-parts can be configured to be
    NO-OP class, which will make that part turned-off in all occurrence.

    Notice: NO_OP will accept any init-args, and ignore them.
    '''
    def __init__(self,*argv,**kws):
        super(NO_OP,self).__init__()
    @staticmethod
    def forward(x,*argv,**kws):
        return x
    forward.__doc__ = __doc__


class InstanceMeanStd(torch.nn.Module):
    '''calculate the mean and std of each sample in a batch.
    This is for pixel-wise statictical feature extractions only.
    For images only.
    '''
    __doc__ = instance_mean_std.__doc__
    def __init__(self,num_features):
        super(self.__class__,self).__init__()
        self.num_features=num_features
    def forward(self,x):
        return instance_mean_std(x,self.num_features)
    forward.__doc__ = __doc__


class InstanceReNorm(torch.nn.Module):
    '''re normalize input with given mean and std-dev value.'''
    __doc__ = instance_renorm.__doc__
    def __init__(self,eps=1e-5):
        super(self.__class__,self).__init__()
        self.eps=eps
        #self.register_buffer(name='eps',tensor=torch.tensor(eps))
    def forward(self,x,mean,std):
        return instance_renorm(x,mean,std,self.eps)
    forward.__doc__ = __doc__

class MaxMinNorm(Clip):
    __doc__ = max_min_norm.__doc__
    def __init__(self,max_or_min,min_or_max,eps=1e-5):
        super(MaxMinNorm,self).__init__(max_or_min,min_or_max)
        self.eps = eps
        #self.register_buffer(name='eps',tensor=eps)
    def forward(self,x):
        return max_min_norm(x,self.maxv,self.minv,self.eps)

class ReLU(torch.nn.ReLU):
    '''activation that accepts any argument and ignores them.
    useful when you want to switch between different activations
    programatically, '''
    def __init__(self,*args,inplace=False,**kws):
        super(ReLU,self).__init__(inplace=inplace)

class ReLUwithGrad(torch.nn.Module):
    '''a relu function with gradient not clipped to zero.
    the gradient w.r.t input is not zero even when the input
    is negative value.

    This RELU output the value like a normal relu max(0,input)
    but it output backward() like a linear function y=x,
    dReLU/dx = 1'''
    def __init__(self,*args,**kws):
        super().__init__()
    def forward(self,x):
        xmirror = x.detach()
        xnegative[x>=0]=0
        return x - xnegative

class Conv2dDepthFullyShared(torch.nn.Conv2d):
    NotImplemented

class Interpolate(torch.nn.Module):
    ''' resizing/scaling multi dimensional tensors
    The modes available for resizing are: `nearest`, `linear` (3D-only),
    `bilinear` (4D-only), `trilinear` (5D-only), `area`

    Args:
        input (Tensor): the input tensor
        size (int or Tuple[int] or Tuple[int, int] or Tuple[int, int, int]):
            output spatial size.
        scale_factor (float or Tuple[float]): multiplier for spatial size. Has to match input size if it is a tuple.
        mode (string): algorithm used for upsampling:
            'nearest' | 'linear' | 'bilinear' | 'trilinear' | 'area'. Default: 'nearest'
        align_corners (bool, optional): if True, the corner pixels of the input
            and output tensors are aligned, and thus preserving the values at
            those pixels. This only has effect when :attr:`mode` is `linear`,
            `bilinear`, or `trilinear`. Default: False

    .. warning::
        With ``align_corners = True``, the linearly interpolating modes
        (`linear`, `bilinear`, and `trilinear`) don't proportionally align the
        output and input pixels, and thus the output values can depend on the
        input size. This was the default behavior for these modes up to version
        0.3.1. Since then, the default behavior is ``align_corners = False``.
        See :class:`~torch.nn.Upsample` for concrete examples on how this
        affects the outputs.
'''
    def __init__(self,size=None,scale_factor=None,mode='bilinear',align_corners=None):
        super().__init__()
        self.__dict__.update(locals())
        if (size is None) and (scale_factor is None):
            print('Torchfun: warning: size and scale_factor are both None, the interpolation parmeters are not specified!')
    def forward(self,x):
        return torch.nn.functional.interpolate(x,
            size=self.size,
            scale_factor=self.scale_factor,
            mode=self.mode,
            align_corners=self.align_corners)


import os,shutil
from glob import glob
from tqdm import tqdm
from sys import argv
import argparse
import imageio,img2pdf
from PIL import Image

parser = argparse.ArgumentParser(description='Image Format Transformer')
parser.add_argument('files', metavar='FILE_PATTERN', type=str, nargs='+',
                        help='files to be transformed, can be like xx/yy/z.jpg, or like xx/yy/*.png')
parser.add_argument('-t','--type',dest='type', type=str, default=None, help='output image type')
parser.add_argument('-d','--dir',dest='dir', type=str, default=None, help='dir to store output images, default: same dir as the input image.')


def img2pdf_wrapper(srcpath,destpath):
    img = Image.open(srcpath)
    with open(destpath,'wb') as f:
        pdf_bytes = img2pdf.convert(img.filename)
        f.write(pdf_bytes)
        f.close()
    img.close()
    return

def imgformat():
    opt=parser.parse_args(argv[1:])
    if not opt.type:
        print('please specify output type by -t or --type possible values: png,jpg,eps,bmp,gif...')
        return 0
    else:
        opt.type = opt.type.lower()

    more_files = []
    files = []
    for fpath in opt.files:
        if '*' in fpath:
            more_files.extend(glob(fpath))
        else:
            files.append(fpath)
    all_fpaths = list(set(more_files+files))
    filenumbers = len(all_fpaths)
    for i,imgpath in enumerate(all_fpaths):
        i = i+1
        srcdir = os.path.dirname(imgpath)
        srcname = os.path.basename(imgpath)
        srctag = '.'.join(srcname.split('.')[:-1])
        srctype = srcname.split('.')[-1].lower()
        destdir = opt.dir or srcdir
        destpath = os.path.join(destdir,srctag+'.'+opt.type)
        if opt.type == srctype: # no need to transform nor read file
            if os.path.abspath(srcdir) != os.path.abspath(destdir): 
                shutil.copy(imgpath,destpath)
                print(i,'/',filenumbers,imgpath,'copy->',destpath)
            else:
                print(i,'/',filenumbers,'skipping',imgpath)
        else: # really need to transform format
            print(i,'/',filenumbers,imgpath,'->',destpath)
            if opt.type == 'pdf':
                img2pdf_wrapper(imgpath,destpath)
            else:
                imageio.imsave(destpath,imageio.imread(imgpath))
    print('finished.')
    return 0

def main():
    return imgformat()

if __name__ == '__main__':
  main()